# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2018 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://github.com/gabylaunay/pyDSA_core


# This file is part of pyDSA_core

# pyDSA_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import numpy as np
import matplotlib.pyplot as plt
from IMTreatment.utils import ProgressCounter
from IMTreatment import TemporalPoints, plotlib as pplt
from .temporalfits import TemporalCircleFits, TemporalSplineFits, \
    TemporalEllipseFits, TemporalCirclesFits, TemporalEllipsesFits
from . import dropedges, dropfit


"""  """

__author__ = "Gaby Launay"
__copyright__ = "Gaby Launay 2017"
__credits__ = ""
__email__ = "gaby.launay@tutanota.com"
__status__ = "Development"


class TemporalDropEdges(TemporalPoints):
    def __init__(self, *args, **kwargs):
        """

        """
        self.baseline = None
        super().__init__(*args, **kwargs)

    def __iter__(self):
        for i in range(len(self.point_sets)):
            yield self.point_sets[i]

    def __getitem__(self, i):
        return self.point_sets[i]

    def fit(self, k=5, s=0.75, verbose=False):
        """
        Compute a spline fitting for the droplets shape.

        Parameters
        ----------
        k : int, optional
            Degree of the smoothing spline.  Must be <= 5.
            Default is k=5.
        s : float, optional
            Smoothing factor between 0 (not smoothed) and 1 (very smoothed)
            Default to 0.75
        """
        return self.fit_spline(k=k, s=s, verbose=verbose)

    def fit_spline(self, k=5, s=0.75, iteration_hook=None, verbose=False):
        """
        Compute a spline fitting for the droplets shape.

        Parameters
        ----------
        k : int, optional
            Degree of the smoothing spline.  Must be <= 5.
            Default is k=5.
        s : float, optional
            Smoothing factor between 0 (not smoothed) and 1 (very smoothed)
            Default to 0.75
        iteration_hook: function
            Hook run at each iterations with the iteration number
            and the total number of iterations planned.
        """
        if verbose:
            pg = ProgressCounter(init_mess="Fitting droplet interfaces",
                                 nmb_max=len(self.point_sets),
                                 name_things='edges', perc_interv=5)
        fits = []
        for i, edge in enumerate(self.point_sets):
            try:
                fits.append(edge.fit_spline(k=k, s=s, verbose=False))
            except Exception:
                fits.append(dropfit.DropFit(baseline=edge.baseline,
                                            x_bounds=edge.x_bounds,
                                            y_bounds=edge.y_bounds))
            if verbose:
                pg.print_progress()
            if iteration_hook is not None:
                iteration_hook(i, len(self.point_sets))
        # return
        tf = TemporalSplineFits(fits=fits, temporaledges=self)
        return tf

    def fit_polyline(self, deg=5, iteration_hook=None, verbose=False):
        """
        Compute a polyline fitting for the droplets shape.

        Parameters
        ----------
        deg : integer
            Degree of the polynomial fitting.
        iteration_hook: function
            Hook run at each iterations with the iteration number
            and the total number of iterations planned.
        """
        if verbose:
            pg = ProgressCounter(init_mess="Fitting droplet interfaces",
                                 nmb_max=len(self.point_sets),
                                 name_things='edges', perc_interv=5)
        fits = []
        for i, edge in enumerate(self.point_sets):
            try:
                fits.append(edge.fit_polyline(deg=deg, verbose=False))
            except Exception:
                fits.append(dropfit.DropFit(baseline=edge.baseline,
                                            x_bounds=edge.x_bounds,
                                            y_bounds=edge.y_bounds))
            if verbose:
                pg.print_progress()
            if iteration_hook is not None:
                iteration_hook(i, len(self.point_sets))
        # return
        tf = TemporalSplineFits(fits=fits, temporaledges=self)
        return tf

    def fit_circle(self, triple_pts=None, iteration_hook=None, verbose=False):
        """
        Fit a circle to the edges.

        Ignore the lower part of the drop if triple points are presents.

        Parameters
        ----------
        iteration_hook: function
            Hook run at each iterations with the iteration number
            and the total number of iterations planned.
        """
        if verbose:
            pg = ProgressCounter(init_mess="Fitting circles to edges",
                                 nmb_max=len(self.point_sets),
                                 name_things='edges',
                                 perc_interv=5)
        fits = []
        for i, edge in enumerate(self.point_sets):
            try:
                fits.append(edge.fit_circle(triple_pts=triple_pts))
            except Exception:
                fits.append(dropfit.DropFit(baseline=edge.baseline,
                                            x_bounds=edge.x_bounds,
                                            y_bounds=edge.y_bounds))
            if verbose:
                pg.print_progress()
            if iteration_hook is not None:
                iteration_hook(i, len(self.point_sets))
        # return
        tf = TemporalCircleFits(fits=fits, temporaledges=self)
        return tf

    def fit_ellipse(self, triple_pts=None, iteration_hook=None, verbose=False):
        """
        Fit an ellipse to the edges.

        Ignore the lower part of the drop if triple points are presents.

        Parameters
        ----------
        iteration_hook: function
            Hook run at each iterations with the iteration number
            and the total number of iterations planned.
        """
        if verbose:
            pg = ProgressCounter(init_mess="Fitting ellipses to edges",
                                 nmb_max=len(self.point_sets),
                                 name_things='edges',
                                 perc_interv=5)
        fits = []
        for i, edge in enumerate(self.point_sets):
            try:
                fits.append(edge.fit_ellipse(triple_pts=triple_pts))
            except Exception:
                fits.append(dropfit.DropFit(baseline=edge.baseline,
                                            x_bounds=edge.x_bounds,
                                            y_bounds=edge.y_bounds))
            if verbose:
                pg.print_progress()
            if iteration_hook is not None:
                iteration_hook(i, len(self.point_sets))
        # return
        tf = TemporalEllipseFits(fits=fits, temporaledges=self)
        return tf

    def fit_ellipses(self, triple_pts=None, iteration_hook=None, verbose=False):
        """
        Fit two ellipses (one on each side) to the edges.

        Ignore the lower part of the drop if triple points are presents.

        Parameters
        ----------
        iteration_hook: function
            Hook run at each iterations with the iteration number
            and the total number of iterations planned.
        """
        if verbose:
            pg = ProgressCounter(init_mess="Fitting ellipses to edges",
                                 nmb_max=len(self.point_sets),
                                 name_things='edges',
                                 perc_interv=5)
        fits = []
        for i, edge in enumerate(self.point_sets):
            try:
                fits.append(edge.fit_ellipses(triple_pts=triple_pts))
            except Exception:
                fits.append(dropfit.DropFit(baseline=edge.baseline,
                                            x_bounds=edge.x_bounds,
                                            y_bounds=edge.y_bounds))
            if verbose:
                pg.print_progress()
            if iteration_hook is not None:
                iteration_hook(i, len(self.point_sets))
        # return
        tf = TemporalEllipsesFits(fits=fits, temporaledges=self)
        return tf

    def fit_circles(self, triple_pts, sigma_max=None, verbose=False,
                    nmb_pass=1, soft_constr=False,
                    iteration_hook=None):
        """
        Fit circles to the edges, cutting them if a triple point is
        present.

        Parameters
        ==========
        sigma_max: number
            If specified, points too far from the fit are iteratively removed
            until:
            std(R) < mean(R)*sigma_max
            With R the radii.
        nmb_pass: positive integer
            If superior to 1, specify the number of pass to make.
            Addintional passes use the previously triple points detected by
            circle fits to more accurately detect the next ones.
        iteration_hook: function
            Hook run at each iterations with the iteration number
            and the total number of iterations planned.
        """
        if verbose:
            pg = ProgressCounter(init_mess="Fitting circles to edges",
                                 nmb_max=len(self.point_sets)*nmb_pass,
                                 name_things='edges',
                                 perc_interv=5)
        # check
        if len(triple_pts) != len(self.point_sets):
            raise Exception('Not the right ner of triple points')
        #  passes
        tf = None
        for i in range(nmb_pass):
            if tf is not None:
                tf.smooth_triple_points('gaussian', size=10)
                triple_pts = tf.get_triple_points()
            fits = []
            for tp, edge in zip(triple_pts, self.point_sets):
                try:
                    fits.append(edge.fit_circles(triple_pts=tp,
                                                 sigma_max=sigma_max,
                                                 soft_constr=soft_constr))
                except Exception:
                    fits.append(dropfit.DropFit(baseline=edge.baseline,
                                                x_bounds=edge.x_bounds,
                                                y_bounds=edge.y_bounds))
                if verbose:
                    pg.print_progress()
                if iteration_hook is not None:
                    iteration_hook(i, len(self.point_sets)*nmb_pass)
            tf = TemporalCirclesFits(fits=fits, temporaledges=self)
        # return
        return tf

    def display(self, *args, **kwargs):
        #
        length = len(self.point_sets)
        displs = []

        # Display points
        for edge in self.point_sets:
            if edge.drop_edges is None:
                edge._separate_drop_edges()
        if self[0].drop_edges is not None:
            x1s = []
            y1s = []
            x2s = []
            y2s = []
            for edge in self.point_sets:
                if edge.drop_edges[0] is None:
                    x1s.append([])
                    x2s.append([])
                    y1s.append([])
                    y2s.append([])
                else:
                    x1s.append(edge.drop_edges[0].y)
                    y1s.append(edge.drop_edges[1].y)
                    x2s.append(edge.drop_edges[2].y)
                    y2s.append(edge.drop_edges[3].y)
            if len(x1s) != length or len(x2s) != length:
                raise Exception()
            db1 = pplt.Displayer(x1s, y1s, color='k', marker="o")
            db2 = pplt.Displayer(x2s, y2s, color='k', marker="o")
            displs.append(db1)
            displs.append(db2)
        # Add button manager
        if len(displs) != 0:
            bm = pplt.ButtonManager(displs)
            return bm

    def display_ridge_evolution(self, tis=None, from_circ_fit=False):
        """
        Display an interactive plot with the ridge height evolution.
        """
        # Get dat
        rh1, rh2 = self.get_ridge_heights(from_circ_fit=from_circ_fit)
        rh = rh1.copy()
        rh[np.isnan(rh1)] = rh2[np.isnan(rh1)]
        filt = np.logical_and(~np.isnan(rh1), ~np.isnan(rh2))
        rh[filt] = (rh1[filt] + rh2[filt])/2
        # Check if there is ridge to display
        if not np.any(filt):
            raise Exception('No ridge detected, cannot display them...')

        # Display images + edges
        plt.figure()
        bmgr = self.display()
        if tis is not None:
            tis.display()
        # Display ridge evolution
        fig2 = plt.figure()
        #     function to navigate through images
        def on_click(event):
            if fig2.canvas.manager.toolbar._active is not None:
                return None
            x = event.xdata
            ind = np.argmin(abs(x - self.times))
            bmgr.ind = ind
            bmgr.update()
        fig2.canvas.mpl_connect('button_press_event', on_click)
        displ2 = pplt.Displayer([[xi] for xi in self.times],
                                [[yi] for yi in rh],
                                data_type="points",
                                color=self[0].colors[2],
                                edgecolors='k')
        bmgr2 = pplt.ButtonManager([displ2])
        plt.plot(self.times, rh1, zorder=0)
        plt.plot(self.times, rh2, zorder=0)
        plt.ylim(ymin=0)
        bmgr.link_to_other_graph(bmgr2)
        return bmgr, bmgr2



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2018 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://github.com/gabylaunay/pyDSA_core

# This file is part of pyDSA_core

# pyDSA_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import cv2
import os
import json
import re
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.widgets import Button, TextBox
import scipy.ndimage as spim
import warnings
from IMTreatment import ScalarField, Profile
from IMTreatment.utils import make_unit
import IMTreatment.plotlib as pplt
from .dropedges import DropEdges
from .baseline import Baseline


"""  """

__author__ = "Gaby Launay"
__copyright__ = "Gaby Launay 2017"
__credits__ = ""
__email__ = "gaby.launay@tutanota.com"
__status__ = "Development"


class Image(ScalarField):
    def __init__(self, filepath=None, cache_infos=True):
        """
        Class representing a greyscale image.
        """
        super().__init__()
        self.baseline = None
        self.colors = pplt.get_color_cycles()
        if filepath is not None:
            self.filepath = os.path.abspath(filepath)
            self.infofile_path = os.path.splitext(self.filepath)[0] + ".info"
            self.cache_infos = cache_infos
        else:
            self.filepath = None
            self.infofile_path = None
            self.cache_infos = False

    def __eq__(self, other):
        if not isinstance(other, Image):
            return False
        if not super().__eq__(other):
            return False
        if not self.baseline == other.baseline:
            return False
        return True

    def display(self, *args, **kwargs):
        displ_baseline = True
        if 'displ_baseline' in kwargs.keys():
            displ_baseline = kwargs.pop('displ_baseline')
        super().display(*args, **kwargs)
        if self.baseline is not None and displ_baseline:
            bx, by = self.baseline.xy
            plt.plot(bx, by, color=self.colors[0], ls='none', marker='o')
            plt.plot(bx, by, color=self.colors[0], ls='-')

    def set_baseline(self, pt1, pt2):
        """
        Set the baseline.

        Parameters
        ==========
        pt1, pt2 : 2x1 arrays of numbers
            Points defining the baseline.
        """
        self.baseline = Baseline([pt1, pt2],
                                 xmax=self.axe_x[-1],
                                 xmin=self.axe_x[0])
        if self.cache_infos:
            self._dump_infos()

    def choose_baseline(self):
        """
        Choose baseline position interactively.

        Select as many points as you want (click on points to delete them),
        and close the figure when done.

        You can adjust the position of the point closer to your mouse with
        arrows, ctrl+arrows and shift+arrows.
        """
        pos = []
        eps = .01*(self.axe_x[-1] - self.axe_x[0])
        dx, dy = self.dx, self.dy
        # get cursor position on click
        fig = plt.figure()
        pts = plt.plot([], marker="o", ls="none", mfc=self.colors[0],
                       mec='k')[0]
        hl_pt = plt.plot([], marker="o", ls="none", mec=self.colors[0],
                         mfc='none',
                         ms=10)[0]
        baseline = plt.plot([], ls="-", color=self.colors[0])[0]
        bs = Baseline()

        def onclick(event):
            # toolbar want the focus !
            if fig.canvas.manager.toolbar._active is not None:
                return None
            # get the position
            xy = [event.xdata, event.ydata]
            diffs = [(xy[0] - xyi[0])**2 + (xy[1] - xyi[1])**2
                     for xyi in pos]
            # check if close to an existing point
            closes = diffs < eps**2
            if np.any(closes):
                for i in np.arange(len(pos) - 1, -1, -1):
                    if closes[i]:
                        del pos[i]
            else:
                pos.append(xy)
            if len(pos) == 0:
                hl_pt.set_data([np.nan, np.nan])
            else:
                hl_pt.set_data(np.array(pos[-1:]).transpose())
            redraw_if_necessary()

        def on_mouse_move(event):
            # toolbar want the focus !
            if fig.canvas.manager.toolbar._active is not None:
                return None
            ind = get_closest_point(event.xdata, event.ydata)
            if ind is not None:
                hl_pt.set_data(np.array(pos[ind:ind+1]).transpose())
                fig.canvas.draw()

        def redraw_if_necessary():
            if len(pos) != 0:
                pts.set_data(np.array(pos).transpose())
                if len(pos) > 1:
                    bs.from_points(pos, xmin=self.axe_x[0],
                                   xmax=self.axe_x[-1])
                    baseline.set_data(bs.xy)
                fig.canvas.draw()

        def get_closest_point(x, y):
            # get the position
            if x is None or y is None:
                return None
            if len(pos) == 0:
                return None
            xy = [x, y]
            diffs = [(xy[0] - xyi[0])**2 + (xy[1] - xyi[1])**2
                     for xyi in pos]
            # check if close to an existing point
            ind = np.argmin(diffs)
            return ind

        def onpress(event):
            x, y = event.xdata, event.ydata
            ind = get_closest_point(x, y)
            if ind is None:
                return None
            if event.key == 'up':
                pos[ind][1] += dy
            elif event.key == 'down':
                pos[ind][1] -= dy
            elif event.key == 'left':
                pos[ind][0] -= dx
            elif event.key == 'right':
                pos[ind][0] += dx
            elif event.key == 'shift+up':
                pos[ind][1] += 10*dy
            elif event.key == 'shift+down':
                pos[ind][1] -= 10*dy
            elif event.key == 'shift+left':
                pos[ind][0] -= 10*dx
            elif event.key == 'shift+right':
                pos[ind][0] += 10*dx
            elif event.key == 'ctrl+up':
                pos[ind][1] += 0.1*dy
            elif event.key == 'ctrl+down':
                pos[ind][1] -= 0.1*dy
            elif event.key == 'ctrl+left':
                pos[ind][0] -= 0.1*dx
            elif event.key == 'ctrl+right':
                pos[ind][0] += 0.1*dx
            # Update indicator
            hl_pt.set_data(np.array(pos[ind:ind+1]).transpose())
            redraw_if_necessary()

        fig.canvas.mpl_connect('button_press_event', onclick)
        fig.canvas.mpl_connect('motion_notify_event', on_mouse_move)
        fig.canvas.mpl_connect('key_press_event', onpress)
        self.display(cmap=plt.cm.binary_r)
        plt.title("Put some points on the baseline."
                  "\nYou can remove points by clicking on it."
                  "\nClose the window when you are happy with the baseline")
        plt.show(block=True)
        if len(pos) < 2:
            if self.baseline is not None:
                return self.baseline
            else:
                return None
        # use linear interpolation to get baseline
        self.baseline = Baseline(pos, xmin=self.axe_x[0],
                                 xmax=self.axe_x[-1])
        if self.cache_infos:
            self._dump_infos()
        return self.baseline

    def scale_interactive(self):
        """
        Scale the Image interactively.
        """
        sc = ScaleChooser(self)
        if self.cache_infos:
            self._dump_infos()
        return sc.ret_values

    def scale(self, scalex=None, scaley=None, scalev=None, inplace=True):
        """
        Scale the Image.

        Parameters
        ----------
        scalex, scaley, scalev : numbers or Unum objects
            Scale for the axis and the values
        """
        if inplace:
            super().scale(scalex=scalex, scaley=scaley, scalev=scalev,
                          inplace=True)
            tmpim = self
        else:
            tmpim = super().scale(scalex=scalex, scaley=scaley, scalev=scalev,
                                  inplace=False)
        if tmpim.baseline is not None:
            tmpim.baseline.scale(scalex=scalex, scaley=scaley)
        if tmpim.cache_infos:
            tmpim._dump_infos()
        return tmpim

    def set_origin(self, x, y):
        """
        Modify the axis in order to place the origin at the given point (x, y)

        Parameters
        ----------
        x : number
        y : number
        """
        super().set_origin(x, y)
        if self.baseline is not None:
            bpt1 = self.baseline.pt1
            bpt2 = self.baseline.pt2
            bpt1[0] -= x
            bpt2[0] -= x
            bpt1[1] -= y
            bpt2[1] -= y
            self.baseline.from_points([bpt1, bpt2])
        if self.cache_infos:
            self._dump_infos()

    def _dump_infos(self):
        # Gather old information if necessary
        if os.path.isfile(self.infofile_path):
            with open(self.infofile_path, 'r') as f:
                dic = json.load(f)
        else:
            dic = {}
        # Update with new values
        unit_x = self.unit_x.strUnit()[1:-1]
        unit_y = self.unit_y.strUnit()[1:-1]
        if self.baseline is not None:
            pt1 = list(self.baseline.pt1)
            pt2 = list(self.baseline.pt2)
        else:
            pt1 = None
            pt2 = None
        new_dic = {"dx": self.dx,
                   "dy": self.dy,
                   "x0": self.axe_x[0],
                   "y0": self.axe_y[0],
                   "baseline_pt1": pt1,
                   "baseline_pt2": pt2,
                   "unit_x": unit_x,
                   "unit_y": unit_y}
        dic.update(new_dic)
        # Write back infos
        with open(self.infofile_path, 'w+') as f:
            json.dump(dic, f)

    def _import_infos(self):
        if not os.path.isfile(self.infofile_path):
            return None
        try:
            with open(self.infofile_path, 'r') as f:
                dic = json.load(f)
        except:
            warnings.warn('Corrupted infofile, reinitializing...')
            os.remove(self.infofile_path)
            return None
        # Update with infos
        dx = dic['dx']
        dy = dic['dy']
        self.scale(scalex=dx/(self.axe_x[1] - self.axe_x[0]),
                   scaley=dy/(self.axe_y[1] - self.axe_y[0]),
                   inplace=True)
        try:  # Compatibility
            x0 = dic['x0']
            y0 = dic['y0']
            self.axe_x += (x0 - self.axe_x[0])
            self.axe_y += (y0 - self.axe_y[0])
        except KeyError:
            pass
        self.unit_x = dic['unit_x']
        self.unit_y = dic['unit_y']
        base1 = dic['baseline_pt1']
        base2 = dic['baseline_pt2']
        if base1 is not None and base2 is not None:
            self.set_baseline(base1, base2)

    def choose_tp_interactive(self):
        """
        Choose triple points interactively.
        """
        sc = TriplePointChooser(self)
        return sc.rh

    def take_measure(self, kind='all'):
        """
        Allow to interactively take a measure on the image.

        Parameters
        ----------
        kind: string
           Kind of measurement to make.
           Can be 'dx', 'dy', 'dl' ot 'all'.
        """
        mt = MeasuringTool(self, kind=kind)
        return mt.measures

    def edge_detection_canny(self, threshold1=None, threshold2=None,
                             base_max_dist=15, size_ratio=.5,
                             nmb_edges=2, ignored_pixels=2,
                             keep_exterior=True,
                             smooth_size=None,
                             dilatation_steps=1,
                             verbose=False, debug=False):
        """
        Perform edge detection using canny edge detection.

        Parameters
        ==========
        threshold1, threshold2: integers
            Thresholds for the Canny edge detection method.
            (By default, inferred from the data histogram)
        base_max_dist: integers
            Maximal distance (in pixel) between the baseline and
            the beginning of the drop edge (default to 15).
        size_ratio: number
            Minimum size of edges, regarding the bigger detected one
            (default to 0.5).
        nmb_edges: integer
            Number of maximum expected edges (default to 2).
        ignored_pixels: integer
            Number of pixels ignored around the baseline
            (default to 2).
            Putting a small value to this allow to avoid
            small surface defects to be taken into account.
        keep_exterior: boolean
            If True (default), only keep the exterior edges.
        smooth_size: number
            If specified, the image is smoothed before
            performing the edge detection.
            (can be useful to put this to 1 to get rid of compression
             artefacts on images).
        dilatation_steps: positive integer
            Number of dilatation/erosion steps.
            Increase this if the drop edges are discontinuous.
        """
        # check for baseline
        if self.baseline is None:
            raise Exception('You should set the baseline first.')
        if not np.isclose(self.dx, self.dy):
            warnings.warn('dx is different than dy, results can be weird...')
        # Get adapted thresholds
        if threshold2 is None or threshold1 is None:
            # hist = self.get_histogram(cum=True,
            #                           bins=int((self.max - self.min)/10))
            mini = 0
            maxi = 255
            hist = cv2.calcHist([self.values], [0], None,
                                [maxi - mini], [mini, maxi])
            hist = np.cumsum(hist[:, 0])
            hist = Profile(np.arange(mini, maxi), hist)
            threshold1 = hist.get_value_position(
                hist.min + (hist.max - hist.min)/2)[0]
            threshold2 = np.max(hist.x)*.5
        # Remove useless part of the image
        tmp_im = self.crop(intervy=[np.min([self.baseline.pt2[1],
                                            self.baseline.pt1[1]]), np.inf],
                           inplace=False)
        # Smooth if asked
        if smooth_size is not None:
            if smooth_size != 0:
                tmp_im.smooth(tos='gaussian', size=smooth_size, inplace=True)
        if verbose:
            plt.figure()
            tmp_im.display()
            plt.title('Initial image')
        #======================================================================
        # Perform Canny detection
        #======================================================================
        im = np.array(tmp_im.values, dtype=np.uint8)
        im_edges = cv2.Canny(image=im, threshold1=threshold1,
                             threshold2=threshold2)
        if verbose:
            plt.figure()
            im = Image()
            im.import_from_arrays(tmp_im.axe_x, tmp_im.axe_y,
                                  im_edges, mask=tmp_im.mask,
                                  unit_x=tmp_im.unit_x, unit_y=tmp_im.unit_y)
            im.display()
            plt.title('Canny edge detection \nwith th1={} and th2={}'
                      .format(threshold1, threshold2))
        #======================================================================
        # Remove points behind the baseline (and too close to)
        #======================================================================
        fun = self.baseline.get_baseline_fun()
        ign_dy = ignored_pixels*self.dy
        max_y = np.max([self.baseline.pt2[1] + ign_dy,
                        self.baseline.pt1[1] + ign_dy])
        for j in range(im_edges.shape[1]):
            y = tmp_im.axe_y[j]
            if y > max_y:
                continue
            for i in range(im_edges.shape[0]):
                x = tmp_im.axe_x[i]
                if y < fun(x) + ign_dy:
                    im_edges[i, j] = 0
        if verbose:
            plt.figure()
            im = Image()
            im.import_from_arrays(tmp_im.axe_x, tmp_im.axe_y,
                                  im_edges, mask=tmp_im.mask,
                                  unit_x=tmp_im.unit_x, unit_y=tmp_im.unit_y)
            im.display()
            plt.title('Removed points under baseline')
        #======================================================================
        # Dilatation / erosion to ensure line continuity
        #======================================================================
        if dilatation_steps > 0:
            im_edges = cv2.dilate(im_edges,
                                  iterations=dilatation_steps,
                                  kernel=np.array([[1, 1, 1],
                                                   [1, 1, 1],
                                                   [1, 1, 1]]))
        # im_edges = spim.binary_erosion(im_edges, iterations=dilatation_steps,
        #                                structure=[[0, 1, 0], [1, 1, 1], [0, 1, 0]])
        if verbose:
            plt.figure()
            im = Image()
            im.import_from_arrays(tmp_im.axe_x, tmp_im.axe_y,
                                  im_edges, mask=tmp_im.mask,
                                  unit_x=tmp_im.unit_x, unit_y=tmp_im.unit_y)
            im.display()
            plt.title('Dilatation / erosion step')
        #======================================================================
        # Keep only the bigger edges
        #======================================================================
        nmb, labels = cv2.connectedComponents(im_edges)
        # labels, nmb = spim.label(im_edges, np.ones((3, 3)))
        nmb_edge = nmb
        dy = self.axe_y[1] - self.axe_y[0]
        # Let only the maximum allowed number of edges
        if np.max(labels) > nmb_edges and not keep_exterior:
            sizes = [np.sum(labels == label)
                     for label in np.arange(1, nmb + 1)]
            indsort = np.argsort(sizes)
            for ind in indsort[:-nmb_edges]:
                im_edges[ind + 1 == labels] = 0
            nmb_edge = nmb_edges
            if verbose:
                plt.figure()
                im = Image()
                im.import_from_arrays(tmp_im.axe_x, tmp_im.axe_y,
                                      im_edges, mask=tmp_im.mask,
                                      unit_x=tmp_im.unit_x,
                                      unit_y=tmp_im.unit_y)
                im.display()
                plt.title('Removed small edges because too numerous')
        if nmb_edge > 1:
            #==================================================================
            # Remove small patches
            #==================================================================
            sizes = [np.sum(labels == label)
                     for label in np.arange(1, nmb + 1)]
            crit_size = np.sort(sizes)[-1]*size_ratio
            for i, size in enumerate(sizes):
                if size < crit_size:
                    im_edges[labels == i+1] = 0
                    labels[labels == i+1] = 0
                    nmb_edge -= 1
            if verbose:
                plt.figure()
                im = Image()
                im.import_from_arrays(tmp_im.axe_x, tmp_im.axe_y,
                                      im_edges, mask=tmp_im.mask,
                                      unit_x=tmp_im.unit_x, unit_y=tmp_im.unit_y)
                im.display()
                plt.title('Removed small edges')
            #==================================================================
            # Remove lines not touching the baseline
            #==================================================================
            if nmb_edge > nmb_edges:
                for i in range(np.max(labels)):
                    # Untested ! (next line)
                    ys = tmp_im.axe_y[np.sum(labels == i+1, axis=0) > 0]
                    if len(ys) == 0:
                        continue
                    min_y = np.min(ys)
                    mdist = base_max_dist*dy
                    if min_y > mdist + np.max([self.baseline.pt1[1],
                                               self.baseline.pt2[1]]):
                        if debug:
                            print(f"Label {i+1} removed because not touching "
                                  "baseline")
                        im_edges[labels == i+1] = 0
                        labels[labels == i+1] = 0
                        nmb_edge -= 1
                if verbose:
                    plt.figure()
                    im = Image()
                    im.import_from_arrays(tmp_im.axe_x, tmp_im.axe_y,
                                          im_edges, mask=tmp_im.mask,
                                          unit_x=tmp_im.unit_x, unit_y=tmp_im.unit_y)
                    im.display()
                    plt.title('Removed edge not touching the baseline')
            #==================================================================
            # Keep only the exterior edges
            #==================================================================
            if nmb_edge > 2 and keep_exterior:
                mean_xs = [np.mean(tmp_im.axe_x[np.sum(labels == i, axis=1)
                                                > 0])
                           for i in np.arange(1, nmb + 1)]
                mean_xs = np.asarray(mean_xs)
                mean_xs[np.isnan(mean_xs)] = np.mean(mean_xs[~np.isnan(mean_xs)])
                mean_xs_indsort = np.argsort(mean_xs)
                for i in mean_xs_indsort[1:-1]:
                    if debug:
                        print(f"Label {i+1} removed because not exterior")
                    im_edges[labels == i+1] = 0
                    labels[labels == i+1] = 0
                    nmb_edge -= 1
            if verbose:
                plt.figure()
                im = Image()
                im.import_from_arrays(tmp_im.axe_x, tmp_im.axe_y,
                                      im_edges, mask=tmp_im.mask,
                                      unit_x=tmp_im.unit_x,
                                      unit_y=tmp_im.unit_y)
                im.display()
                plt.title('Removed the interior edged')
        #======================================================================
        # Check and Return
        #======================================================================
        # Get points coordinates
        xs, ys = np.where(im_edges)
        axx = tmp_im.axe_x
        axy = tmp_im.axe_y
        xs = [axx[x] for x in xs]
        ys = [axy[y] for y in ys]
        xys = list(zip(xs, ys))
        # Check if there is something remaining
        if len(xys) < 10:
            if verbose:
                plt.show(block=False)
            raise Exception("Didn't found any drop here...")
        if verbose:
            plt.figure()
            self.display()
            xys = np.array(xys)
            plt.plot(xys[:, 0], xys[:, 1], ".k")
            plt.title('Initial image + edge points')
        edge = DropEdges(xy=xys, im=self, type="canny")
        return edge

    def edge_detection_contour(self, nmb_edges=2, ignored_pixels=2,
                               level=0.5, size_ratio=.5,
                               verbose=False):
        """
        Perform edge detection using contour.

        Parameters
        ==========
        nmb_edges: integer
            Number of maximum expected edges (default to 2).
        size_ratio: number
            Minimum size of edges, regarding the bigger detected one
            (default to 0.5).
        level: number
            Normalized level of the drop contour.
            Should be between 0 (black) and 1 (white).
            Default to 0.5.
        ignored_pixels: integer
            Number of pixels ignored around the baseline
            (default to 2).
            Putting a small value to this allow to avoid
            small surface defects to be taken into account.
        """
        if self.baseline is None:
            raise Exception('You should set the baseline first.')
        if self.dx != self.dy:
            warnings.warn('dx is different than dy, results can be weird...')
        # Get level
        level = self.min + level*(self.max - self.min)
        # Get contour
        import skimage.measure as skim
        contour = skim.find_contours(self.values, level)
        # put in the right units
        for cont in contour:
            cont[:, 0] *= self.dx
            cont[:, 0] += self.axe_x[0]
            cont[:, 1] *= self.dy
            cont[:, 1] += self.axe_y[0]
        # Remove contours below the baseline (or too close to)
        fun = self.baseline.get_baseline_fun()
        ign_dy = ignored_pixels*self.dy
        max_y = np.max([self.baseline.pt2[1] + ign_dy,
                        self.baseline.pt1[1] + ign_dy])
        new_contour = []
        for cont in contour:
            new_cont = []
            for xy in cont:
                if xy[1] > max_y + ign_dy:
                    new_cont.append(xy)
                elif xy[1] > fun(xy[0]) + ign_dy:
                    new_cont.append(xy)
            if len(new_cont) != 0:
                new_contour.append(new_cont)
        contour = new_contour
        # Keep only the bigger edges
        contour = sorted(contour, key=lambda x: -len(x))
        contour = contour[0:nmb_edges]
        threshold = len(contour[0])*size_ratio
        contour = [cont for cont in contour if len(cont) > threshold]
        # concatenate
        xys = np.concatenate(contour)
        # Check if there is something remaining
        if len(xys) < 10:
            raise Exception("Didn't found any drop here...")
        # return
        edge = DropEdges(xy=xys, im=self, type="contour")
        return edge

    edge_detection = edge_detection_canny


class ScaleChooser(object):
    def __init__(self, im):
        """

        """
        self.fig = plt.figure()
        self.pos = []
        self.im = im
        self.eps = .01*(self.im.axe_x[-1] - self.im.axe_x[0])
        self.pts = plt.plot([], marker="o", ls="none", mec='w', mfc='k')[0]
        self.axplot = plt.gca()
        self.ret_values = [None, None]
        self.real_dist = None
        self.real_unit = None
        # Display
        self.im.display(cmap=plt.cm.binary_r)
        plt.title("Scaling step:\n"
                  "Choose two points separated by a known distance.")
        # Connect click event on graph
        self.fig.canvas.mpl_connect('button_press_event', self.onclick)
        # Add ok button
        button_kwargs = {"color": "w", "hovercolor": [.5]*3}
        self.axok = mpl.axes.Axes(self.fig, [0.88, 0.02, 0.1, 0.05])
        self.bnok = Button(self.axok, 'Done', **button_kwargs)
        self.bnok.on_clicked(self.done)        # set up the apsect
        self.fig.add_axes(self.axok)
        # Add text
        self.axtxt = mpl.axes.Axes(self.fig, [0.68, 0.02, 0.1, 0.05])
        self.btntxt = TextBox(self.axtxt,
                              "Distance ('0.7mm' for example):", "")
        self.btntxt.on_submit(self.on_submit)
        self.fig.add_axes(self.axtxt)
        # show the plot
        plt.show(block=True)

    def on_submit(self, txt):
        match = re.match(r'\s*([0-9.]+)\s*(.*)\s*', txt)
        self.real_dist = float(match.groups()[0])
        self.real_unit = match.groups()[1]

    def onclick(self, event):
        # toolbar want the focus !
        if self.fig.canvas.manager.toolbar._active is not None:
            return None
        # Do nothing if not above an axe
        if event.inaxes != self.axplot:
            return None
        # get the position
        xy = [event.xdata, event.ydata]
        diffs = [(xy[0] - xyi[0])**2 + (xy[1] - xyi[1])**2
                 for xyi in self.pos]
        # check if close to an existing point
        closes = diffs < self.eps**2
        if np.any(closes):
            for i in np.arange(len(self.pos) - 1, -1, -1):
                if closes[i]:
                    del self.pos[i]
        elif len(self.pos) >= 2:
            pass
        else:
            self.pos.append(xy)
        # redraw
        if len(self.pos) != 0:
            self.pts.set_data(np.array(self.pos).transpose())
        else:
            self.pts.set_data(np.empty((2, 2)))
        self.fig.canvas.draw()

    def done(self, event):
        if len(self.pos) != 2:
            return None, None
        actual_width = ((self.pos[0][0] - self.pos[1][0])**2 +
                        (self.pos[0][1] - self.pos[1][1])**2)**.5
        # Getting wanted length and unity
        wanted_width = self.real_dist
        wanted_unity = make_unit(self.real_unit)
        # Compute and set the scale
        scale = wanted_width/actual_width
        self.im.scale(scalex=scale, scaley=scale, inplace=True)
        self.im.unit_x = wanted_unity
        self.im.unit_y = wanted_unity
        self.ret_values = scale, wanted_unity
        plt.close(self.fig)


class TriplePointChooser(object):
    def __init__(self, im):
        """
        """
        self.fig = plt.figure()
        if im.baseline is None:
            raise Exception()
        self.basefun = im.baseline.get_baseline_fun()
        self.pos = []
        self.basepos = []
        self.rh = []
        self.im = im
        self.eps = .01*(self.im.axe_x[-1] - self.im.axe_x[0])
        self.pts = plt.plot([], marker="o", ls=":", color='k',
                            mec='w', mfc='k')[0]
        self.texts = []
        self.axplot = plt.gca()
        # Display
        self.im.display(cmap=plt.cm.binary_r)
        plt.title("Click on the triple point positions.")
        # Connect click event on graph
        self.fig.canvas.mpl_connect('button_press_event', self.onclick)
        # show the plot
        plt.show(block=True)

    def onclick(self, event):
        # toolbar want the focus !
        if self.fig.canvas.manager.toolbar._active is not None:
            return None
        # Do nothing if not above an axe
        if event.inaxes != self.axplot:
            return None
        # get the position
        xy = [event.xdata, event.ydata]
        diffs = [(xy[0] - xyi[0])**2 + (xy[1] - xyi[1])**2
                 for xyi in self.pos]
        # check if close to an existing point
        closes = diffs < self.eps**2
        if np.any(closes):
            for i in np.arange(len(self.pos) - 1, -1, -1):
                if closes[i]:
                    del self.pos[i]
                    del self.rh[i]
                    del self.basepos[i]
                    self.texts[i].remove()
                    del self.texts[i]
        else:
            self.pos.append(xy)
            self.basepos.append(self.im.baseline.get_projection_to_baseline(xy))
            self.rh.append(self.im.baseline.get_distance_to_baseline(xy))
            if self.im.unit_y.strUnit() != "[]":
                unit = self.im.unit_y.strUnit()[1:-1]
            else:
                unit = ""
            self.texts.append(plt.text(self.pos[-1][0],
                                       self.pos[-1][1] + self.eps,
                                       f"h={self.rh[-1]:.2f}{unit}",
                                       horizontalalignment='center',
                                       verticalalignment='baseline'))
        # redraw
        if len(self.pos) != 0:
            pos = np.concatenate([[self.pos[i],
                                   self.basepos[i],
                                   [np.nan, np.nan]]
                                  for i in range(len(self.pos))])
            self.pts.set_data(np.array(pos).transpose())
        else:
            self.pts.set_data(np.empty((2, 2)))
        self.fig.canvas.draw()


class MeasuringTool(object):
    def __init__(self, im, kind='all'):
        """
        """
        self.im = im
        self.kind = kind
        #
        self.fig = plt.figure()
        self.im.display(cmap=plt.cm.binary_r)
        self.axplot = plt.gca()
        #
        self.pts_couples = [[]]
        self.measures = {'dx': [], 'dy': [], 'dl': []}
        self.plots = []
        self.texts = []
        self.plot_args = {'marker': 'o', 'ls': '-'}
        #
        self.eps = .01*(self.im.axe_x[-1] - self.im.axe_x[0])
        # Connect click event on graph
        self.fig.canvas.mpl_connect('button_press_event', self.onclick)
        # show the plot
        plt.show(block=True)

    def onclick(self, event):
        # toolbar want the focus !
        if self.fig.canvas.manager.toolbar._active is not None:
            return None
        # Do nothing if not above an axe
        if event.inaxes != self.axplot:
            return None
        # get the position
        xy = [event.xdata, event.ydata]
        # if fist point of a couple
        if len(self.pts_couples[-1]) == 0:
            self.pts_couples[-1].append(xy)
            plot = plt.plot([xy[0]], [xy[1]], **self.plot_args)[0]
            self.plots.append(plot)
        # if second point of a couple
        else:
            self.pts_couples[-1].append(xy)
            xs = [self.pts_couples[-1][0][0], xy[0]]
            ys = [self.pts_couples[-1][0][1], xy[1]]
            self.plots[-1].set_data([xs, ys])
            self.pts_couples.append([])
            # add text
            if self.im.unit_x != self.im.unit_y:
                unit = ""
            else:
                unit = self.im.unit_x.strUnit()
            text = ""
            print(f"Points {len(self.pts_couples)-1}:")
            dx = xs[1] - xs[0]
            dy = ys[1] - ys[0]
            dl = (dx**2 + dy**2)**.5
            if self.kind in ['dx', 'all']:
                print(f"dx={dx} {unit}")
                text += f'dx={dx:.2f} {unit}\n'
            if self.kind in ['dy', 'all']:
                print(f"dy={dy} {unit}")
                text += f'dy={dy:.2f} {unit}\n'
            if self.kind in ['dl', 'all']:
                print(f"dl={dl} {unit}")
                text += f'dl={dl:.2f} {unit}\n'
            text = text.rstrip('\n')
            self.texts.append(plt.text(np.mean(xs),
                                       np.mean(ys),
                                       text,
                                       horizontalalignment='center',
                                       verticalalignment='baseline',
                                       bbox=dict(facecolor='white', alpha=0.5)))
            # record
            self.measures['dx'].append(dx)
            self.measures['dy'].append(dy)
            self.measures['dl'].append(dl)
        self.fig.canvas.draw()



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2018 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://github.com/gabylaunay/pyDSA_core

# This file is part of pyDSA_core

# pyDSA_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import numpy as np
import matplotlib.pyplot as plt
import unum
import copy


"""  """

__author__ = "Gaby Launay"
__copyright__ = "Gaby Launay 2017"
__credits__ = ""
__email__ = "gaby.launay@tutanota.com"
__status__ = "Development"


class Baseline(object):
    def __init__(self, pts=None, xmin=None, xmax=None):
        """

        """
        self.pt1 = None
        self.pt2 = None
        self.xy = []
        self.coefs = []
        self.tilt_angle = None
        if pts is not None:
            self.from_points(pts=pts, xmin=xmin, xmax=xmax)

    def __eq__(self, other):
        if not isinstance(other, Baseline):
            return False
        for arg in ['pt1', 'pt2', 'xy', 'coefs', 'tilt_angle']:
            try:
                getattr(self, arg)[0]
                if not np.allclose(getattr(self, arg), getattr(other, arg)):
                    return False
            except (TypeError, IndexError):
                if not np.isclose(getattr(self, arg), getattr(other, arg)):
                    return False
        return True

    def copy(self):
        return copy.deepcopy(self)

    def from_points(self, pts, xmin=None, xmax=None):
        if len(pts) == 2 and ((xmin is None and xmax is None)
                              or (xmin == pts[0][0] and xmax == pts[1][0])):
            self.pt1 = np.array(pts[0], dtype=float)
            self.pt2 = np.array(pts[1], dtype=float)
        else:
            self.pt1, self.pt2 = self.get_baseline_from_points(pts,
                                                               xmin=xmin,
                                                               xmax=xmax)
        # get coefs
        self.xy = [[self.pt1[0], self.pt2[0]], [self.pt1[1], self.pt2[1]]]
        slope = (self.pt2[1] - self.pt1[1])/(self.pt2[0] - self.pt1[0])
        intercept = self.pt2[1] - slope*self.pt2[0]
        self.coefs = [slope, intercept]
        self.tilt_angle = np.arctan(self.coefs[0])

    @staticmethod
    def get_baseline_from_points(pts, xmin=None, xmax=None):
        pos = np.array(pts)
        a, b = np.polyfit(pos[:, 0], pos[:, 1], 1)
        if xmin is None:
            xmin = np.min(pos[:, 0])
        if xmax is None:
            xmax = np.max(pos[:, 0])
        res = np.array([[xmin, a*xmin + b], [xmax, a*xmax + b]],
                       dtype=float)
        return res

    def get_baseline_fun(self, along_y=False):
        a, b = self.coefs
        # check if flat:
        if abs(a) < abs(b)*1e-10 and along_y:
            raise Exception("baseline is flat")
        if along_y:
            def fun(y):
                return (y - b)/a
        else:
            def fun(x):
                return a*x + b
        return fun

    def get_projection_to_baseline(self, xy):
        """
        Return the projection of the given point on the baseline.
        """
        vec1 = np.array([self.pt2[0] - self.pt1[0], self.pt2[1] - self.pt1[1]])
        vec2 = np.array([xy[0] - self.pt1[0], xy[1] - self.pt1[1]])
        vec3 = np.dot(vec1, vec2)/np.dot(vec1, vec1)*vec1
        return np.array([vec3[0] + self.pt1[0], vec3[1] + self.pt1[1]])

    def get_distance_to_baseline(self, xy):
        """
        Return the distance between the given point and the baseline/
        """
        bspos = self.get_projection_to_baseline(xy)
        return np.linalg.norm(xy - bspos)

    def scale(self, scalex=None, scaley=None):
        """
        Scale the baseline.
        """
        if isinstance(scalex, unum.Unum):
            scalex = float(scalex.asNumber())
        if isinstance(scaley, unum.Unum):
            scaley = float(scaley.asNumber())
        # print(scalex)
        # print(scaley)
        pt1 = self.pt1
        pt2 = self.pt2
        if scalex is not None:
            pt1[0] *= scalex
            pt2[0] *= scalex
        if scaley is not None:
            pt1[1] *= scaley
            pt2[1] *= scaley
        if (scalex is not None) or (scaley is not None):
            self.from_points([pt1, pt2])

    def set_origin(self, xo, yo):
        """
        Change the origin of the baseline to the new point.
        """
        x = np.array([self.pt1[0], self.pt2[0]])
        y = np.array([self.pt1[1], self.pt2[1]])
        new_x = x - xo
        new_y = y - yo
        self.from_points([[new_x[0], new_y[0]],
                          [new_x[1], new_y[1]]])

    def rotate(self, angle):
        """
        Rotate the baseline.

        Parameters
        ----------
        angle: number
            Rotation angle in radian.
        """
        # Checks
        x = np.array([self.pt1[0], self.pt2[0]])
        y = np.array([self.pt1[1], self.pt2[1]])
        new_x = x*np.cos(angle) - y*np.sin(angle)
        new_y = y*np.cos(angle) + x*np.sin(angle)
        self.from_points([[new_x[0], new_y[0]],
                          [new_x[1], new_y[1]]])

    def display(self, x0=None, xf=None, *args, **kwargs):
        if x0 is None:
            x0 = self.xy[0][0]
            y0 = self.xy[1][0]
        else:
            fun = self.get_baseline_fun()
            y0 = fun(x0)
        if xf is None:
            xf = self.xy[0][1]
            yf = self.xy[1][1]
        else:
            fun = self.get_baseline_fun()
            yf = fun(xf)
        plt.plot([x0, xf], [y0, yf], *args, **kwargs)



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2018 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://github.com/gabylaunay/pyDSA_core

# This file is part of pyDSA_core

# pyDSA_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate as spint
import scipy.optimize as spopt
from IMTreatment import Points, Profile
import IMTreatment.plotlib as pplt
from .helpers import fit_circle, fit_ellipse, get_ellipse_points
from .dropfit import DropCircleFit, DropEllipseFit, DropSplineFit, \
    DropCirclesFit, DropEllipsesFit


"""  """

__author__ = "Gaby Launay"
__copyright__ = "Gaby Launay 2017"
__credits__ = ""
__email__ = "gaby.launay@tutanota.com"
__status__ = "Development"


def dummy_function(x):
    try:
        return np.array([np.nan]*len(x))
    except TypeError:
        return np.nan


def poly_function_from_coefs(coefs):
    def fun(x):
        nmb_coefs = len(coefs)
        x = np.asarray(x)
        return np.sum([coefs[i]*x**(nmb_coefs - 1 - i)
                       for i in range(nmb_coefs)], axis=0)
    return fun


class DropEdges(Points):
    def __init__(self, xy, im, type):
        """
        """
        super().__init__(xy=xy,
                         unit_x=im.unit_x,
                         unit_y=im.unit_y)
        self._type = type
        self.drop_edges = None
        self.baseline = im.baseline
        self.im_axe_x = im.axe_x
        self.im_axe_y = im.axe_y
        self.x_bounds = [im.axe_x[0], im.axe_x[-1]]
        self.y_bounds = [im.axe_y[0], im.axe_y[-1]]
        self.im_dx = im.dx
        self.im_dy = im.dy
        self.colors = pplt.get_color_cycles()

    def _separate_drop_edges(self):
        """
        Separate the two sides of the drop.
        """
        # check if edges are present...
        if len(self.xy) == 0:
            self.drop_edges = [None]*4
            return [None]*4
        # Get cut position
        ind_sort = np.argsort(self.xy[:, 0])
        xs = self.xy[:, 0][ind_sort]
        ys = self.xy[:, 1][ind_sort]
        dxs = xs[1::] - xs[0:-1]
        dxs_sorted = np.sort(dxs)
        if dxs_sorted[-1] > 10*dxs_sorted[-2]:
            # ind of the needle center (needle)
            ind_cut = np.argmax(dxs) + 1
            x_cut = xs[ind_cut]
        else:
            # ind of the higher point (no needle)
            ind_cut = np.argmax(ys)
            x_cut = xs[ind_cut]
        # Separate according to cut position
        xs1 = xs[xs < x_cut]
        xs2 = xs[xs > x_cut]
        ys1 = ys[xs < x_cut]
        ys2 = ys[xs > x_cut]
        # # Ensure only one y per x (Use hash table for optimization)
        new_y1 = np.sort(list(set(ys1)))
        dic = {y1: 0 for y1 in new_y1}
        nmb = {y1: 0 for y1 in new_y1}
        for i in range(len(ys1)):
            dic[ys1[i]] += xs1[i]
            nmb[ys1[i]] += 1
        new_x1 = [dic[y1]/nmb[y1] for y1 in new_y1]
        new_y2 = np.sort(list(set(ys2)))
        dic = {y2: 0 for y2 in new_y2}
        nmb = {y2: 0 for y2 in new_y2}
        for i in range(len(ys2)):
            dic[ys2[i]] += xs2[i]
            nmb[ys2[i]] += 1
        new_x2 = [dic[y2]/nmb[y2] for y2 in new_y2]
        # smooth to avoid stepping
        # Profile also automatically sort along y
        de1 = Profile(new_y1, new_x1)
        de2 = Profile(new_y2, new_x2)
        de1.smooth(tos='gaussian', size=1, inplace=True)
        de2.smooth(tos='gaussian', size=1, inplace=True)
        # parametrize
        if len(de1) != 0:
            t1s = np.cumsum(((de1.x[1:] - de1.x[0:-1])**2
                             + (de1.y[1:] - de1.y[0:-1])**2)**.5)
            t1s = np.concatenate(([0], t1s))
            t1s /= t1s[-1]
        else:
            t1s = []
        if len(de2) != 0:
            t2s = np.cumsum(((de2.x[1:] - de2.x[0:-1])**2
                             + (de2.y[1:] - de2.y[0:-1])**2)**.5)
            t2s = np.concatenate(([0], t2s))
            t2s /= t2s[-1]
        else:
            t2s = []
        dex1 = Profile(t1s, de1.y, unit_x="", unit_y=self.unit_x)
        dex2 = Profile(t2s, de2.y, unit_x="", unit_y=self.unit_x)
        dey1 = Profile(t1s, de1.x, unit_x="", unit_y=self.unit_y)
        dey2 = Profile(t2s, de2.x, unit_x="", unit_y=self.unit_y)
        # evenlify
        if len(dex1) != 0:
            dtx1 = self.im_dx/abs(dex1.y[-1] - dex1.y[0])
            dty1 = self.im_dy/abs(dey1.y[-1] - dey1.y[0])
            dt1 = np.min([dtx1, dty1])
            dex1 = dex1.evenly_space(dx=dt1)
            dey1 = dey1.evenly_space(dx=dt1)
        if len(dex2) != 0:
            dtx2 = self.im_dx/abs(dex2.y[-1] - dex2.y[0])
            dty2 = self.im_dy/abs(dey2.y[-1] - dey2.y[0])
            dt2 = np.min([dtx2, dty2])
            dex2 = dex2.evenly_space(dx=dt2)
            dey2 = dey2.evenly_space(dx=dt2)
        # store
        self.drop_edges = (dex1, dey1, dex2, dey2)
        return (dex1, dey1, dex2, dey2)

    def rotate(self, angle, inplace=False):
        """
        Rotate the edge.

        Parameters
        ----------
        angle: number
            Rotation angle in radian.
        """
        # Checks
        if inplace:
            tmpp = self
        else:
            tmpp = self.copy()
        super(DropEdges, tmpp).rotate(angle, inplace=True)
        if tmpp.drop_edges is not None:
            for de in tmpp.drop_edges:
                de.rotate(angle=angle, inplace=True)
        if tmpp.baseline is not None:
            tmpp.baseline.rotate(angle=angle)
        tmpp.edges_fits = None
        tmpp.triple_pts = None
        tmpp.circle_fits = None
        tmpp.circle_triple_pts = None
        return tmpp

    def fit(self, k=5, s=.75, verbose=False):
        """
        Compute a spline fitting for the droplet shape.

        Parameters
        ----------
        k : int, optional
            Degree of the smoothing spline.  Must be <= 5.
            Default is k=5.
        s : float, optional
            Smoothing factor between 0 (not smoothed) and 1 (very smoothed)
            Default to 0.75

        Returns:
        --------
        fit: DropFit object
            .
        """
        return self.fit_spline(k=k, s=s, verbose=verbose)

    def fit_spline(self, k=5, s=.75, verbose=False):
        """
        Compute a spline fitting for the droplet shape.

        Parameters
        ----------
        k : int, optional
            Degree of the smoothing spline.  Must be <= 5.
            Default is k=5.
        s : float, optional
            Smoothing factor between 0 (not smoothed) and 1 (very smoothed)
            Default to 0.75

        Returns:
        --------
        fit: DropFit object
            .
        """
        # Chec if edges are present
        if self.drop_edges is None:
            self._separate_drop_edges()
        if self.drop_edges[0] is None and self.drop_edges[2] is None:
            dsf = DropSplineFit(x_bounds=[np.nan, np.nan],
                                y_bounds=[np.nan, np.nan],
                                baseline=self.baseline,
                                fits=[dummy_function, dummy_function])
            return dsf
        # Prepare drop edge for interpolation
        # TODO: Find a more efficient fitting
        dex1, dey1, dex2, dey2 = self.drop_edges
        spline1 = None
        spline2 = None
        # Don't fit if no edge
        if len(dex1) == 0:
            spline1 = dummy_function
        if len(dex2) == 0:
            spline2 = dummy_function
        # spline interpolation
        min_smooth_carac_len = np.max([self.im_dx, self.im_dy])/6
        max_smooth_carac_len = min_smooth_carac_len*2
        max_len = np.max([len(dex1), len(dex2)])
        min_smooth_fact = max_len*min_smooth_carac_len**2
        max_smooth_fact = max_len*max_smooth_carac_len**2
        s = max_smooth_fact - (max_smooth_fact - min_smooth_fact)*s
        max_smooth_carac_len = np.max([self.im_axe_x[-1] - self.im_axe_x[0],
                                       self.im_axe_y[-1] - self.im_axe_y[0]])
        if verbose:
            print("Used 's={}'".format(s))
        if spline1 is None:
            try:
                spx1 = spint.UnivariateSpline(dex1.x, dex1.y, k=k, s=s)
                spy1 = spint.UnivariateSpline(dey1.x, dey1.y, k=k, s=s)
                spline1 = [spx1, spy1]
            except:
                spline1 = (dummy_function, dummy_function)
                if verbose:
                    print("Fitting failed for edge number one")
        if spline2 is None:
            try:
                spx2 = spint.UnivariateSpline(dex2.x, dex2.y, k=k, s=s)
                spy2 = spint.UnivariateSpline(dey2.x, dey2.y, k=k, s=s)
                spline2 = [spx2, spy2]
            except:
                spline2 = (dummy_function, dummy_function)
                if verbose:
                    print("Fitting failed for edge number one")
        # Verbose if necessary
        if verbose:
            t = np.linspace(0, 1, 1000)
            plt.figure()
            plt.plot(dex1.y, dey1.y, "o")
            plt.plot(spline1[0](t), spline1[1](t), 'r')
            plt.plot(dex2.y, dey2.y, "o")
            plt.plot(spline2[0](t), spline2[1](t), 'r')
            plt.axis('equal')
            plt.show()
        # return
        x_bounds = [np.min(self.xy[:, 0]), np.max(self.xy[:, 0])]
        y_bounds = [np.min(self.xy[:, 1]), np.max(self.xy[:, 1])]
        dsf = DropSplineFit(x_bounds=x_bounds,
                            y_bounds=y_bounds,
                            baseline=self.baseline,
                            fits=[spline1, spline2])
        return dsf

    def fit_polyline(self, deg=5, verbose=False):
        """
        Fit the drop edge with a poly line.

        Parameters
        ----------
        deg: integer
            Polynomial degree (default to 5).

        Returns
        -------
        fit: DropFit object
            .
        """
        if self.drop_edges is None:
            self._separate_drop_edges()
        if self.drop_edges[0] is None and self.drop_edges[2] is None:
            dsf = DropSplineFit(x_bounds=[np.nan, np.nan],
                                y_bounds=[np.nan, np.nan],
                                baseline=self.baseline,
                                fits=[dummy_function, dummy_function])
            return dsf
        # Prepare drop edge for interpolation
        # TODO: Find a more efficient fitting
        dex1, dey1, dex2, dey2 = self.drop_edges
        polyline1 = None
        polyline2 = None
        # Don't fit if no edge
        if len(dey1) == 0:
            polyline1 = (dummy_function, dummy_function)
        if len(dey2) == 0:
            polyline2 = (dummy_function, dummy_function)
        # fit
        if polyline1 is None:
            try:
                psx1 = np.polyfit(dex1.x, dex1.y, deg)
                psy1 = np.polyfit(dey1.x, dey1.y, deg)
                polyline1 = [poly_function_from_coefs(psx1),
                             poly_function_from_coefs(psy1)]
            except:
                polyline1 = (dummy_function, dummy_function)
                if verbose:
                    print("Fitting failed for edge number one")
        if polyline2 is None:
            try:
                psx2 = np.polyfit(dex2.x, dex2.y, deg)
                psy2 = np.polyfit(dey2.x, dey2.y, deg)
                polyline2 = [poly_function_from_coefs(psx2),
                             poly_function_from_coefs(psy2)]
            except:
                polyline2 = (dummy_function, dummy_function)
                if verbose:
                    print("Fitting failed for edge number two")
        # Return
        x_bounds = [np.min(self.xy[:, 0]), np.max(self.xy[:, 0])]
        y_bounds = [np.min(self.xy[:, 1]), np.max(self.xy[:, 1])]
        dsf = DropSplineFit(x_bounds=x_bounds,
                            y_bounds=y_bounds,
                            baseline=self.baseline,
                            fits=[polyline1, polyline2])
        return dsf

    def fit_ellipse(self, triple_pts=None, verbose=False):
        """
        Fit the drop edges with an ellipse.

        Remove points under the triple points if possible.

        Returns:
        --------
        fit: DropFit object
            .
        """
        if self.drop_edges is None:
            self._separate_drop_edges()
        dex1, dey1, dex2, dey2 = self.drop_edges
        if dex1 is None or dex2 is None:
            des = DropEllipseFit(xyc=[np.nan, np.nan], R1=np.nan,
                                 R2=np.nan, theta=np.nan,
                                 baseline=self.baseline,
                                 x_bounds=[np.nan, np.nan],
                                 y_bounds=[np.nan, np.nan])
            return des
        xs1 = dex1.y
        ys1 = dey1.y
        xs2 = dex2.y
        ys2 = dey2.y
        if triple_pts is not None:
            tp1 = triple_pts[0]
            tp2 = triple_pts[1]
            if not np.isnan(tp1[1]):
                filt1 = ys1 > tp1[1]
                xs1 = xs1[filt1]
                ys1 = ys1[filt1]
            if not np.isnan(tp2[1]):
                filt2 = ys2 > tp2[1]
                xs2 = xs2[filt2]
                ys2 = ys2[filt2]
            if verbose:
                plt.figure()
                self.display()
                plt.figure()
                plt.plot(xs1, ys1)
                plt.plot(xs2, ys2)
                plt.axis('equal')
                plt.show()
        xs = np.concatenate((xs1, xs2))
        ys = np.concatenate((ys1, ys2))
        (xc, yc), R1, R2, theta = fit_ellipse(xs, ys)
        if verbose:
            plt.figure()
            self.display()
            thetas = np.linspace(0, np.pi*2, 100)
            xs = xc + R1*np.cos(thetas)
            ys = yc + R2*np.sin(thetas)
            new_xs = xc + np.cos(theta)*(xs - xc) - np.sin(theta)*(ys - yc)
            new_ys = yc + np.cos(theta)*(ys - yc) + np.sin(theta)*(xs - xc)
            plt.figure()
            self.display()
            plt.plot(xc, yc, 'ok')
            plt.plot(new_xs, new_ys)
            plt.show()
        # Return
        x_bounds = [np.min(self.xy[:, 0]), np.max(self.xy[:, 0])]
        y_bounds = [np.min(self.xy[:, 1]), np.max(self.xy[:, 1])]
        des = DropEllipseFit(xyc=[xc, yc], R1=R1, R2=R2, theta=theta,
                             baseline=self.baseline,
                             x_bounds=x_bounds,
                             y_bounds=y_bounds)
        return des

    def fit_ellipses(self, triple_pts=None, verbose=False):
        """
        Fit the drop edges with two ellipses (one on each sides).

        Remove points under the triple points if possible.

        Returns:
        --------
        fit: DropFit object
            .
        """
        if self.drop_edges is None:
            self._separate_drop_edges()
        dex1, dey1, dex2, dey2 = self.drop_edges
        if dex1 is None or dex2 is None:
            des = DropEllipseFit(xyc=[np.nan, np.nan], R1=np.nan,
                                 R2=np.nan, theta=np.nan,
                                 baseline=self.baseline,
                                 x_bounds=[np.nan, np.nan],
                                 y_bounds=[np.nan, np.nan])
            return des
        xs1 = dex1.y
        ys1 = dey1.y
        xs2 = dex2.y
        ys2 = dey2.y
        if triple_pts is not None:
            tp1 = triple_pts[0]
            tp2 = triple_pts[1]
            if not np.isnan(tp1[1]):
                filt1 = ys1 > tp1[1]
                xs1 = xs1[filt1]
                ys1 = ys1[filt1]
            if not np.isnan(tp2[1]):
                filt2 = ys2 > tp2[1]
                xs2 = xs2[filt2]
                ys2 = ys2[filt2]
            if verbose:
                plt.figure()
                self.display()
                plt.figure()
                plt.plot(xs1, ys1)
                plt.plot(xs2, ys2)
                plt.axis('equal')
                plt.show()
        fit1 = fit_ellipse(xs1, ys1)
        fit2 = fit_ellipse(xs2, ys2)
        if verbose:
            plt.figure()
            self.display()
            thetas = np.linspace(0, np.pi*2, 100)
            for fit in [fit1, fit2]:
                (xc, yc), R1, R2, theta = fit
                xs = xc + R1*np.cos(thetas)
                ys = yc + R2*np.sin(thetas)
                new_xs = xc + np.cos(theta)*(xs - xc) - np.sin(theta)*(ys - yc)
                new_ys = yc + np.cos(theta)*(ys - yc) + np.sin(theta)*(xs - xc)
                plt.figure()
                self.display()
                plt.plot(xc, yc, 'ok')
                plt.plot(new_xs, new_ys)
            plt.show()
        # Return
        x_bounds = [np.min(self.xy[:, 0]), np.max(self.xy[:, 0])]
        y_bounds = [np.min(self.xy[:, 1]), np.max(self.xy[:, 1])]
        des = DropEllipsesFit(xyc1=fit1[0], R1a=fit1[1], R1b=fit1[2],
                              theta1=fit1[3],
                              xyc2=fit2[0], R2a=fit2[1], R2b=fit2[2],
                              theta2=fit2[3],
                              baseline=self.baseline,
                              x_bounds=x_bounds,
                              y_bounds=y_bounds)
        return des

    def fit_LY(self, interp_res=100, method=None, verbose=False):
        """
        Fit the drop edges with the Laplace-young equations.

        Returns:
        --------
        fit: DropFit object
            .
        """
        raise Exception('Not implemented (properly) yes')
        # get data
        z1_min = triple_pts[0][1]
        z1 = np.linspace(z1_min,
                         np.max(self.xy[:, 1]),
                         interp_res)
        z2_min = self.triple_pts[1][1]
        z2 = np.linspace(z2_min,
                         np.max(self.xy[:, 1]),
                         interp_res)
        dz = z1[1] - z1[0]
        r1 = abs(self.edges_fits[0](z1) - self.get_drop_position())
        r2 = abs(self.edges_fits[1](z2) - self.get_drop_position())
        if verbose:
            plt.figure()
            self.display()
            plt.figure()
            plt.plot(r1, z1, label="edge1")
            plt.plot(r2, z2, label="edge2")
            plt.xlabel('r')
            plt.ylabel('z')
            plt.legend()
            plt.show()
        # Get gradients
        r1 = np.asarray(r1)
        r1p = np.gradient(r1, dz)
        r1pp = np.gradient(r1p, dz)
        z1p = np.gradient(z1, r1)
        z1pp = np.gradient(z1p, r1)
        if verbose:
            plt.figure()
            plt.plot(r1, z1/np.mean(z1), label="z1")
            plt.plot(r1, z1p/np.mean(z1p), label="z1p")
            plt.plot(r1, z1pp/np.mean(z1pp), label="z1pp")
            plt.xlabel('r')
            plt.ylabel('z')
            plt.legend()
            plt.show()
        # Get curvature
        # R1_inv = r1pp/(1 + r1p**2)**(3/2)
        # R2_inv = 1/(r1*(1 + r1p**2)**(1/2))
        R1_inv = (z1p/(r1*(1 + z1p**2)**.5))
        R2_inv = (z1pp/(1 + z1p**2)**(3/2))
        if verbose:
            R1 = 1/(z1p/(r1*(1 + z1p**2)**.5))
            R2 = 1/(z1pp/(1 + z1p**2)**(3/2))
            plt.figure()
            plt.plot(r1, z1)
            plt.xlabel('r')
            plt.ylabel('z')
            plt.figure()
            plt.plot(R1, z1, label="R1")
            plt.plot(R2, z1, label="R2")
            plt.xlabel('r')
            plt.ylabel('z')
            plt.legend()
            plt.show()
        # fit for right side
        def minfun(args):
            A, B = args
            res = R1_inv + R2_inv - (A + B*z1)
            return np.sum(res**2)
        res1 = spopt.minimize(minfun, [1, 1], method=method)
        if not res1.success:
            raise Exception('Fitting sigma failed:\n{}'.format(res1.message))
        A, B = res1.x
        if verbose:
            plt.figure()
            ls = R1_inv + R2_inv
            plt.plot(ls, z1, label="left equation side")
            rs = A + B*z1
            plt.plot(rs, z1, label="left equation side")
            plt.xlabel('r')
            plt.ylabel('z')
            plt.legend()
            plt.show()
        # Fit for left side (Curvature radius)
        def minfun(args, dz, z):
            r1 = args
            z1p = np.gradient(z1, r1)
            z1pp = np.gradient(z1p, r1)
            R1_inv = (z1p/(r1*(1 + z1p**2)**.5))
            R2_inv = (z1pp/(1 + z1p**2)**(3/2))
            res = (R1_inv + R2_inv) - (A + B*z1)
            return np.sum(res**2)
        res2 = spopt.minimize(minfun, r1, (dz, z1), method=method)
        if not res2.success:
            raise Exception('Fitting drop edge failed:\n{}'
                            .format(res2.message))
        new_r1 = res2.x
        if verbose:
            plt.figure()
            plt.plot(r1, z1, label="Spline fitting")
            plt.plot(new_r1, z1, label="LY fitting")
            plt.legend()
            plt.show()
        # TODO: Need to return a DropFit object !
        return z1, - new_r1 + self.get_drop_position()

    def fit_circle(self, triple_pts=None, verbose=False):
        """
        Fit a circle to the edges.

        Ignore the lower part of the drop if triple points are presents.
        """
        # get data
        if self.drop_edges is None:
            self._separate_drop_edges()
        dex1, dey1, dex2, dey2 = self.drop_edges
        if dex1 is None or dex2 is None:
            dcf = DropCircleFit(xyc=[np.nan, np.nan], R=np.nan,
                                baseline=self.baseline,
                                x_bounds=[np.nan, np.nan],
                                y_bounds=[np.nan, np.nan])
            return dcf
        xs1 = dex1.y
        ys1 = dey1.y
        xs2 = dex2.y
        ys2 = dey2.y
        if triple_pts is not None:
            if not np.any(np.isnan(triple_pts[0])):
                ytp = triple_pts[0][1]
                filt = ys1 > ytp
                xs1 = xs1[filt]
                ys1 = ys1[filt]
            if not np.any(np.isnan(triple_pts[1])):
                ytp = triple_pts[1][1]
                filt = ys2 > ytp
                xs2 = xs2[filt]
                ys2 = ys2[filt]
        xs = np.concatenate((xs1, xs2))
        ys = np.concatenate((ys1, ys2))
        # Fit circles
        c, R = fit_circle(xs, ys)
        self.circle_fits = [[c, R]]
        # verbose
        if verbose:
            plt.plot(xs, ys, '.k')
            plt.plot(c[0], c[1], "ok")
            theta = np.linspace(0, 2*np.pi, 100)
            tmpxs = c[0] + R*np.cos(theta)
            tmpys = c[1] + R*np.sin(theta)
            plt.plot(tmpxs, tmpys, "-k")
            plt.axis('image')
            plt.show()
        # return
        x_bounds = [np.min(self.xy[:, 0]), np.max(self.xy[:, 0])]
        y_bounds = [np.min(self.xy[:, 1]), np.max(self.xy[:, 1])]
        dcf = DropCircleFit(xyc=c, R=R, baseline=self.baseline,
                            x_bounds=x_bounds, y_bounds=y_bounds)
        return dcf

    def fit_circles(self, triple_pts, sigma_max=None, soft_constr=False,
                    verbose=False):
        """
        Fit circles to the edges, cutting them if a triple point is
        present.

        If a triple point is present, the ridge fits are made to be tangent
        to the drop and to the baseline.

        Parameters
        ==========
        sigma_max: number
            If specified, points too far from the fit are iteratively removed
            until:
            std(R) < mean(R)*sigma_max
            With R the radii.

        """
        # get data
        tp1 = triple_pts[0]
        tp2 = triple_pts[1]
        if self.drop_edges is None:
            self._separate_drop_edges()
        dex1, dey1, dex2, dey2 = self.drop_edges
        # check
        if dex1 is None or dex2 is None:
            dcf = DropCirclesFit(xyc=[[np.nan, np.nan],
                                      [np.nan, np.nan],
                                      [np.nan, np.nan]],
                                 Rs=[np.nan, np.nan, np.nan],
                                 baseline=self.baseline,
                                 triple_pts=[[np.nan, np.nan],
                                             [np.nan, np.nan]],
                                 x_bounds=[np.nan, np.nan],
                                 y_bounds=[np.nan, np.nan])
            return dcf
        #
        xs1 = dex1.y
        ys1 = dey1.y
        xs2 = dex2.y
        ys2 = dey2.y
        # reorder
        if ys1[0] > ys1[-1]:
            ys1 = ys1[::-1]
            xs1 = xs1[::-1]
        if ys2[0] > ys2[-1]:
            ys2 = ys2[::-1]
            xs2 = xs2[::-1]
        # separate at the triple point
        ind_sep1 = np.where(ys1 > tp1[1])[0][0]
        ind_sep2 = np.where(ys2 > tp2[1])[0][0]
        xs_o1 = xs1[0:ind_sep1]
        ys_o1 = ys1[0:ind_sep1]
        xs_o2 = xs2[0:ind_sep2]
        ys_o2 = ys2[0:ind_sep2]
        xs_d = np.concatenate((xs1[ind_sep1::], xs2[ind_sep2::][::-1]))
        ys_d = np.concatenate((ys1[ind_sep1::], ys2[ind_sep2::][::-1]))
        # Fit circles
        # (use the middle circle fit to enhance the ridge fits)
        c_d, Rd = fit_circle(xs_d, ys_d,
                             sigma_max=sigma_max)
        c_o1, R1 = fit_circle(xs_o1, ys_o1, self.baseline,
                              sigma_max=sigma_max,
                              tangent_circ=(c_d, Rd),
                              soft_constr=soft_constr)
        c_o2, R2 = fit_circle(xs_o2, ys_o2, self.baseline,
                              sigma_max=sigma_max,
                              tangent_circ=(c_d, Rd),
                              soft_constr=soft_constr)
        # check for abnormal values
        if R1 > Rd or R2 > Rd or c_o1[0] > c_d[0] or c_o2[0] < c_d[0]:
            return None
        # Get triple points based from circles
        xtp1 = (c_o1[0]*Rd + c_d[0]*R1)/(Rd + R1)
        ytp1 = (c_o1[1]*Rd + c_d[1]*R1)/(Rd + R1)
        xtp2 = (c_o2[0]*Rd + c_d[0]*R2)/(Rd + R2)
        ytp2 = (c_o2[1]*Rd + c_d[1]*R2)/(Rd + R2)
        circle_triple_pts = [[xtp1, ytp1], [xtp2, ytp2]]
        # Return
        x_bounds = [np.min(self.xy[:, 0]), np.max(self.xy[:, 0])]
        y_bounds = [np.min(self.xy[:, 1]), np.max(self.xy[:, 1])]
        dcf = DropCirclesFit(xyc=[c_d, c_o1, c_o2],
                             Rs=[Rd, R1, R2],
                             baseline=self.baseline,
                             triple_pts=circle_triple_pts,
                             x_bounds=x_bounds,
                             y_bounds=y_bounds)
        return dcf

    def display(self, *args, **kwargs):
        """
        """
        if self.drop_edges is not None:
            dex1, dey1, dex2, dey2 = self.drop_edges
            xs1 = dex1.y
            ys1 = dey1.y
            xs2 = dex2.y
            ys2 = dey2.y
            plt.plot(xs1, ys1, color='k', marker='o')
            plt.plot(xs2, ys2, color='k', marker='o')
        else:
            plt.plot(self.xy[:, 0], self.xy[:, 1], color='k', marker='o')
        plt.axis('equal')
        plt.gca().set_adjustable('box')
        # Display baseline
        if self.baseline is not None:
            x0 = np.min(self.xy[:, 0])
            xf = np.max(self.xy[:, 0])
            x0 -= np.abs(xf - x0)*.1
            xf += np.abs(xf - x0)*.1
            self.baseline.display(x0, xf, color=self.colors[0])



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2018 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://github.com/gabylaunay/pyDSA_core

# This file is part of pyDSA_core

# pyDSA_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


"""  """

__author__ = "Gaby Launay"
__copyright__ = "Gaby Launay 2017"
__credits__ = ""
__email__ = "gaby.launay@tutanota.com"
__status__ = "Development"

import numpy as np
import matplotlib.pyplot as plt
import scipy.optimize as spopt
from IMTreatment import Profile


def fit_YL(s, x, z, delta_rho, gamma_0=40e-3):
    """
    Fit the Young-Laplace equation to the given points, returning the
    surface tension.

    Parameters
    ----------
    s, x, z: arrays of numbers
        Parametrized interface of the pendant drop.
        s = 0 should correspond to the drop apex
    delta_rho: number
        Difference between the two fluid densities
    gamma_0: number
        Guess for the surface tension (helps for numerical convergence)

    Returns
    -------
    gamma: number
        Surface tension.

    Notes
    -----
    Inspired by https://www.dataphysics-instruments.com/knowledge/
    understanding-interfaces/pendant-drop-method/
    """
    # sort (just in case)
    indsort = np.argsort(s)
    s = s[indsort]
    x = x[indsort]
    z = z[indsort]
    # Angle between the horizontal and the local tangent of the surface
    theta1 = np.arcsin(np.gradient(x, s))
    theta2 = np.arccos(np.gradient(z, s))
    filter_theta1 = np.isnan(theta1)
    filter_theta2 = np.isnan(theta2)
    theta1[filter_theta1] = theta2[filter_theta1]
    theta2[filter_theta2] = theta1[filter_theta2]
    theta = np.mean([theta1, theta2], axis=0)
    theta[np.isnan(theta)] = 0
    print(theta1)
    print(theta2)
    print(theta)
    # Radius of curvature at the apex
    zp = np.gradient(z[0:5], s[0:5])
    zpp = np.gradient(zp, s[0:5])[0]
    zp = zp[0]
    xp = np.gradient(x[0:5], s[0:5])
    xpp = np.gradient(xp, s[0:5])[0]
    xp = xp[0]
    R0 = np.abs((xp**2 + zp**2)**(3/2)/(xp*zpp - zp*xpp))
    #
    thetap = np.gradient(theta, s)[1::]
    K1 = -np.sin(theta[1::])/x[1::] + 2/R0
    g = 9.81
    def minfun(gamma):
        return thetap - K1 - np.abs(delta_rho*g*z[1::]/gamma)
    gamma, res = spopt.leastsq(minfun, (gamma_0, ))
    if res > 4:
        raise Exception(res)
    return gamma




s = np.linspace(0, 1, 100)
x = s
z = 1 - (1 - s**2)**.5
dists = ((x[1::] - x[0:-1])**2 + (z[1::] - z[0:-1])**2)**.5
cumdists = np.cumsum(np.concatenate(([0], dists)))
s = cumdists/cumdists[-1]
x = Profile(s, x)
x.evenly_space(inplace=True)
z = Profile(s, z)
z.evenly_space(inplace=True)
s = z.x
z = z.y
x = x.y
plt.figure()
plt.plot(s, x, 'o')
plt.figure()
plt.plot(s, z, 'o')
plt.figure()
plt.plot(x, z, 'o')

gamma = fit_YL(s, x, z, 1000 - 1.25)



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2018 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://github.com/gabylaunay/pyDSA_core

# This file is part of pyDSA_core

# pyDSA_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import numpy as np
import warnings
import matplotlib.pyplot as plt
import scipy.optimize as spopt
import scipy.misc as spmisc
import IMTreatment.plotlib as pplt
from . import helpers


__author__ = "Gaby Launay"
__copyright__ = "Gaby Launay 2017"
__credits__ = ""
__email__ = "gaby.launay@tutanota.com"
__status__ = "Development"

# TODO: remove tests fot self.fits existence: they always exist !
# TODO: recheck all the docstrings


class DropFit(object):
    def __init__(self, baseline, x_bounds, y_bounds):
        """
        A droplet edges fitting.

        Parameters
        ----------
        im: Image object
            Image at the origin of the fit
        """
        self.baseline = baseline
        self.thetas = None
        self.fits = None
        self.triple_pts = None
        self.thetas_triple = None
        self.x_bounds = x_bounds
        self.y_bounds = y_bounds
        self.colors = pplt.get_color_cycles()

    def display(self):
        """
        """
        plt.axis('equal')
        plt.gca().set_adjustable('box')
        # Display baseline
        if self.baseline is not None:
            x0 = self.x_bounds[0]
            xf = self.x_bounds[-1]
            x0 -= np.abs(xf - x0)*.1
            xf += np.abs(xf - x0)*.1
            self.baseline.display(x0, xf, color=self.colors[0])

    def _get_angle_display_lines(self):
        bs_angle = self.baseline.tilt_angle*180/np.pi
        lines = []
        length = (self.y_bounds[-1] - self.y_bounds[0])/3
        if self.thetas is not None:
            theta1 = (self.thetas[0] + bs_angle)/180*np.pi
            theta2 = (self.thetas[1] + bs_angle)/180*np.pi
            xy_inter = self._get_inters_base_fit()
            y1 = xy_inter[0][1]
            y2 = xy_inter[1][1]
            x1 = xy_inter[0][0]
            x2 = xy_inter[1][0]
            lines.append([[x1, x1 + length*np.cos(theta1)],
                          [y1, y1 + length*np.sin(theta1)]])
            lines.append([[x2, x2 + length*np.cos(theta2)],
                          [y2, y2 + length*np.sin(theta2)]])
        else:
            lines.append([[np.nan, np.nan],
                          [np.nan, np.nan]])
            lines.append([[np.nan, np.nan],
                          [np.nan, np.nan]])
        if self.thetas_triple is not None:
            # contact angle with triple points
            theta1 = (self.thetas_triple[0] + bs_angle)/180*np.pi
            theta2 = (self.thetas_triple[1] + bs_angle)/180*np.pi
            xy_inter = self.triple_pts
            y1 = xy_inter[0][1]
            y2 = xy_inter[1][1]
            x1 = xy_inter[0][0]
            x2 = xy_inter[1][0]
            lines.append([[x1, x1 + length*np.cos(theta1)],
                          [y1, y1 + length*np.sin(theta1)]])
            lines.append([[x2, x2 + length*np.cos(theta2)],
                          [y2, y2 + length*np.sin(theta2)]])
        else:
            lines.append([[np.nan, np.nan],
                          [np.nan, np.nan]])
            lines.append([[np.nan, np.nan],
                          [np.nan, np.nan]])
        return lines

    def get_fit_as_points(self, resolution=100):
        """
        Return a representation of the fit as point coordinates.
        """
        return np.array([[np.nan], [np.nan]])

    def get_drop_center(self):
        raise NotImplementedError("Not implemented yet")

    def get_drop_position(self):
        """ Return the position of the droplet edges. """
        try:
            return self._get_inters_base_fit()
        except:
            return np.array([[np.nan, np.nan], [np.nan, np.nan]])

    def get_base_diameter(self):
        """
        Return the base diameter.
        """
        pt1, pt2 = self._get_inters_base_fit()
        diam = ((pt1[0] - pt2[0])**2 + (pt1[1] - pt2[1])**2)**0.5
        return diam

    def get_drop_height(self):
        """ Return the height of the droplet center. """
        xyc = self.get_drop_center()
        hb = self.baseline.get_projection_to_baseline(xyc)[1]
        hmax = np.max(self.y_bounds)
        return hmax - hb

    def get_drop_area(self):
        raise NotImplementedError("Not implemented yet")

    def get_drop_volume(self):
        raise NotImplementedError("Not implemented yet")

    def get_ridge_height(self):
        """
        Return the ridge height.
        """
        pts = self.triple_pts
        if pts is None:
            return np.nan, np.nan
        heights = [self.baseline.get_distance_to_baseline(pt)
                   for pt in pts]
        return np.array(heights)


class DropSplineFit(DropFit):
    def __init__(self, fits, x_bounds, y_bounds, baseline):
        """
        A droplet edges spline fitting.

        Parameters
        ----------
        im: Image object
            Image at the origin of the fit
        fits: List of 2 functions
            Spline fit for the two drop edges.
        """
        super().__init__(baseline=baseline,
                         x_bounds=x_bounds,
                         y_bounds=y_bounds)
        self.fits = fits

    def _detect_triple_points_as_x_minima(self, edge_number, verbose):
        fit = self.fits[edge_number]
        t = np.linspace(0, 1, 100)
        dt = 1/100
        def zerofun(t):
            dxdt = spmisc.derivative(fit[0], t, dx=dt, n=1, order=3)
            dydt = spmisc.derivative(fit[1], t, dx=dt, n=1, order=3)
            dxdy = dxdt*(1/dydt)
            return dxdy
        def dzerofun(t):
            return spmisc.derivative(zerofun, t, dx=dt, n=1, order=5)
        t0 = t[0]
        deriv = zerofun(t)
        deriv_sign = deriv[1:]*deriv[:-1]
        inds = np.where(deriv_sign < 0)[0] + 1
        # If no x minima
        if len(inds) == 0:
            if verbose:
                plt.figure()
                plt.plot(t, zerofun(t), 'o-')
                plt.xlabel('y')
                plt.xlabel('dy/dx')
                plt.title('x-minima method failed: no minima')
                plt.grid()
            return None
        # Choose first point with right curvature sign
        for indi in inds:
            curv = dzerofun(t[indi])
            if ((edge_number == 0 and curv < 0) or
                (edge_number == 1 and curv > 0)):
                ind = indi
                break
        # Find the accurate position
        t0 = spopt.brentq(zerofun, t[ind-1], t[ind])
        # verbose
        if verbose:
            plt.figure()
            plt.plot(t, zerofun(t), 'o-')
            plt.xlabel('y')
            plt.xlabel('dy/dx')
            plt.axvline(t[ind - 1], ls='--', color='k')
            plt.axvline(t[ind], ls='--', color='k')
            plt.title('x-minima method')
            plt.grid()
        return [fit[0](t0), fit[1](t0)]

    def _detect_triple_points_as_curvature_change(self, edge_number, verbose):
        fit = self.fits[edge_number]
        t = np.linspace(0, 1, 100)
        dt = 1/100
        def zerofun(t):
            dxdt = spmisc.derivative(fit[0], t, dx=dt, n=1, order=3)
            dydt = spmisc.derivative(fit[1], t, dx=dt, n=1, order=3)
            dxdy = dxdt*(1/dydt)
            dxdt2 = spmisc.derivative(fit[0], t, dx=dt, n=2, order=3)
            dydt2 = spmisc.derivative(fit[1], t, dx=dt, n=2, order=3)
            dxdy2 = dxdt2*(1/dydt2)
            return dxdy2/(1 + dxdy**2)**(3/2)
        def dzerofun(t):
            return spmisc.derivative(zerofun, t, dx=dt, n=1, order=5)
        if verbose:
            plt.figure()
            plt.plot(t, zerofun(t), 'o-')
            plt.xlabel('y')
            plt.ylabel('d^2y/dx^2')
            plt.title('curvature method')
            plt.grid()
        # Get the triple point iteratively
        while True:
            t0 = self._get_curvature_root(t=t, zerofun=zerofun,
                                          verbose=verbose)
            if t0 is None:
                return None
            # check if the triple point is curvature coherent,
            # else, find the next one
            deriv = dzerofun(t0)
            if (edge_number == 0 and deriv < 0) or (edge_number == 1
                                                    and deriv > 0):
                t = t[t > t0]
            else:
                break
            # sanity check
            if len(t) == 0:
                if verbose:
                    warnings.warn('Cannot find a decent triple point')
                return None
        # Ok, good to go
        if verbose:
            plt.figure()
            plt.plot(t, zerofun(t), 'o-')
            plt.plot(t0, 0, "ok")
            plt.xlabel('y')
            plt.xlabel('d^2y/dx^2')
            plt.grid()
        return [fit[0](t0), fit[1](t0)]

    def _get_curvature_root(self, t, zerofun, verbose):
        try:
            # get last point with opposite sign
            t0 = t[0]
            tf = t[-1]
            while True:
                if zerofun(t0)*zerofun(tf) > 0 and \
                    abs(t0 - tf)/((t0 + tf)/2) > 0.01:
                    tf -= (tf - t0)*1/10
                else:
                    break
            # get exact position
            t0 = spopt.brentq(zerofun, t0, tf)
            return t0
        except (RuntimeError, ValueError) as m:
            if verbose:
                warnings.warn('Cannot find a triple point here.'
                              '\nYou should try a different fitting.'
                              '\nError message:{}'.format(m))
            return None

    def detect_triple_points(self, verbose=False, use_x_minima=False):
        """
        Compute the triple points (water, oil and air interfaces) position.

        Parameters
        ==========
        use_x_minima: boolean
            If True, try to define the triple point as the minimal x values and
            fall back to the curvature method if necessary.
            (default to False).

        Returns
        =======
        tripl_pts: 2x2 array of numbers
           Position of the triple points for each edge ([pt1, pt2])
        """
        # Checks
        if self.fits is None:
            raise Exception("You should computing fitting first with 'fit()'")
        tripl_pts = [None, None]
        for i in [0, 1]:
            # Try the x minima method
            if use_x_minima:
                tp = self._detect_triple_points_as_x_minima(edge_number=i,
                                                            verbose=verbose)
                if tp is None:
                    tripl_pts[i] = [np.nan, np.nan]
                    if verbose:
                        print("Use of the x minima failed")
                else:
                    tripl_pts[i] = tp
                    continue
            # Try the curvature change method
            tp = self._detect_triple_points_as_curvature_change(
                edge_number=i,
                verbose=verbose)
            if tp is None:
                tripl_pts[i] = [np.nan, np.nan]
                if verbose:
                    print("Use of the curvature change failed")
            else:
                tripl_pts[i] = tp
        # Store and return
        if verbose:
            plt.show()
        self.triple_pts = tripl_pts
        return tripl_pts

    def _get_inters_base_fit(self, verbose=False):
        flat = False
        try:
            bfun = self.baseline.get_baseline_fun(along_y=True)
        except Exception:
            flat = True
        xys = []
        for i in range(2):
            sfunx, sfuny = self.fits[i]
            if np.isnan(sfunx(0)):
                x_inter = np.nan
                y_inter = np.nan
            elif flat:
                y_inter = self.baseline.pt1[1]
                t_inter = spopt.fsolve(lambda t: y_inter - sfuny(t), 0.5)[0]
                x_inter = sfunx(t_inter)
            else:
                t_inter = spopt.fsolve(lambda t: bfun(sfuny(t)) - sfunx(t),
                                       0.5)[0]
                x_inter = sfunx(t_inter)
                y_inter = sfuny(t_inter)
            xys.append([x_inter, y_inter])
        if verbose:
            x = np.linspace(self.baseline.pt1[0], self.baseline.pt2[0], 100)
            bfun = self.baseline.get_baseline_fun()
            plt.figure()
            plt.plot([xys[0][0], xys[1][0]], [xys[0][1], xys[1][1]], "ok",
                     label="intersection")
            plt.plot(x, bfun(x), label="base")
            t = np.linspace(0, 1, 100)
            plt.plot(sfunx(t), sfuny(t), label="fit")
            plt.legend()
            plt.show()
        return xys

    def compute_contact_angle(self, verbose=False):
        """
        Compute the contact angles.

        Returns
        =======
        thetas : 2x1 array of numbers
           Contact angles in °
        """
        bs_angle = self.baseline.tilt_angle*180/np.pi
        # Compute base contact angle
        xy_inter = self._get_inters_base_fit()
        self.thetas = self._compute_fitting_angle_at_pts(xy_inter)
        # correct regarding baseline angle
        self.thetas -= bs_angle
        # Compute triple point contact angle
        if self.triple_pts is not None:
            xy_tri = self.triple_pts
            self.thetas_triple = self._compute_fitting_angle_at_pts(xy_tri)
            # correct regardin baseline angle
            self.thetas_triple -= bs_angle
        # display if asked
        if verbose:
            self.display()

    def _compute_fitting_angle_at_pts(self, pts):
        thetas = []
        for i in range(2):
            x_inter, y_inter = pts[i]
            sfunx, sfuny = self.fits[i]
            t_inter = spopt.fsolve(lambda t: y_inter - sfuny(t), 0.5)
            # Get gradient
            dt = 0.01
            derivx = spmisc.derivative(sfunx, t_inter, dx=dt)
            derivy = spmisc.derivative(sfuny, t_inter, dx=dt)
            theta = np.arctan2(derivy, derivx)[0]
            theta = theta % (2*np.pi)
            thetas.append(theta/np.pi*180)
        return np.array(thetas)

    def get_fit_as_points(self, resolution=100):
        """
        Return a representation of the fit as point coordinates.
        """
        (fit1x, fit1y), (fit2x, fit2y) = self.fits
        t = np.linspace(0, 1, resolution)
        x1 = fit1x(t)
        y1 = fit1y(t)
        x2 = fit2x(t)
        y2 = fit2y(t)
        xs = np.concatenate((x1, [np.nan], x2))
        ys = np.concatenate((y1, [np.nan], y2))
        pts = [xs, ys]
        return pts

    def display(self, displ_fits=True, displ_ca=True,
                displ_tp=True, *args, **kwargs):
        """
        """
        super().display()
        # Display fits
        if self.fits is not None and displ_fits:
            t = np.linspace(0, 1, 1000)
            x1 = self.fits[0][0](t)
            y1 = self.fits[0][1](t)
            x2 = self.fits[1][0](t)
            y2 = self.fits[1][1](t)
            plt.plot(x1, y1, color=self.colors[1])
            plt.plot(x2, y2, color=self.colors[1])
        # Display contact angles
        if displ_ca:
            lines = self._get_angle_display_lines()
            for line in lines:
                plt.plot(line[0], line[1],
                         color=self.colors[0])
                plt.plot(line[0][0], line[0][1],
                         color=self.colors[0])
        # Display triple points
        if self.triple_pts is not None and displ_tp:
            for i in [0, 1]:
                plt.plot(self.triple_pts[i][0],
                         self.triple_pts[i][1],
                         marker="o",
                         color=self.colors[4])

    def get_drop_center(self):
        """ Return the center of the droplet 2D projection. """
        t = np.linspace(0, 1, 1000)
        xs = np.concatenate((self.fits[0][0](t), self.fits[1][0](t)[::-1]))
        ys = np.concatenate((self.fits[0][1](t), self.fits[1][1](t)[::-1]))
        # shoelace
        # Credit to Elfego Ruiz-Gutierrez
        A = self.get_drop_area()
        z = xs * np.roll(ys, 1) - ys * np.roll(xs, 1)
        xx = xs + np.roll(xs, 1)
        yy = ys + np.roll(ys, 1)
        center = np.r_[np.dot(xx, z), np.dot(yy, z)]/A/6.0
        return center

    def get_drop_area(self):
        """ Return the droplet 2D projection area. """
        t = np.linspace(0, 1, 1000)
        xs = np.concatenate((self.fits[0][0](t), self.fits[1][0](t)[::-1]))
        ys = np.concatenate((self.fits[0][1](t), self.fits[1][1](t)[::-1]))
        # Shoelace method
        area = 0.5*np.abs(np.dot(xs, np.roll(ys, 1))
                          - np.dot(ys, np.roll(xs, 1)))
        return area

    def get_drop_volume(self):
        raise Exception('Cannot get the drop volume from a spline fit')


class DropCircleFit(DropFit):
    def __init__(self, xyc, R, baseline, x_bounds, y_bounds):
        super().__init__(baseline=baseline,
                         x_bounds=x_bounds,
                         y_bounds=y_bounds)
        self.fits = [xyc, R]

    def _get_inters_base_fit(self):
        """
        """
        # getting intersection points
        # from: http://mathworld.wolfram.com/Circle-LineIntersection.html
        (xc, yc), r = self.fits
        xbas1 = self.baseline.pt1[0] - xc
        xbas2 = self.baseline.pt2[0] - xc
        ybas1 = self.baseline.pt1[1] - yc
        ybas2 = self.baseline.pt2[1] - yc
        dx = xbas2 - xbas1
        dy = ybas2 - ybas1
        dr = (dx**2 + dy**2)**.5
        D = xbas1*ybas2 - xbas2*ybas1
        #
        if dy == 0:
            sign_dy = 1
        else:
            sign_dy = np.sign(dy)
        x1 = (D*dy + sign_dy*dx*(r**2*dr**2 - D**2)**.5)/dr**2
        x2 = (D*dy - sign_dy*dx*(r**2*dr**2 - D**2)**.5)/dr**2
        y1 = (-D*dx + abs(dy)*(r**2*dr**2 - D**2)**.5)/dr**2
        y2 = (-D*dx - abs(dy)*(r**2*dr**2 - D**2)**.5)/dr**2
        #
        x1 += xc
        x2 += xc
        y1 += yc
        y2 += yc
        return [[x1, y1], [x2, y2]]

    def compute_contact_angle(self, verbose=False):
        """
        Compute the contact angles.

        Returns
        =======
        thetas : 2x1 array of numbers
           Contact angles in °
        """
        bs_angle = self.baseline.tilt_angle*180/np.pi
        # Compute circle fits contact angles
        (xc, yc), R = self.fits
        pt1, pt2 = self._get_inters_base_fit()
        thetas = [- np.pi/2 + np.arctan2((yc - pt1[1]), (xc - pt1[0])),
                  np.pi/2 + np.arctan2((yc - pt2[1]), (xc - pt2[0]))]
        # Be sure to be in the right side of the baseline
        thetas[0] = thetas[0]
        thetas[1] = thetas[1]
        beta = self.baseline.tilt_angle
        if (beta - thetas[0]) % (2*np.pi) < np.pi:
            thetas[0] = (thetas[0] + np.pi) % (2*np.pi)
        if (beta - thetas[1]) % (2*np.pi) < np.pi:
            thetas[1] = (thetas[1] + np.pi) % (2*np.pi)
        # Convert to degree
        self.thetas = np.array(thetas)*180/np.pi
        # correct regarding the baseline angle
        self.thetas -= bs_angle
        # display if asked
        if verbose:
            self.display()

    def get_drop_center(self):
        """
        Return the center of the drop.
        """
        xc, yc = self.fits[0]
        # r = self.fits[1]
        # h = yc - self.baseline.get_projection_to_baseline([xc, yc])[1]
        # Need to be in the baseline referential...
        print("Warning: not in the baseline referential")
        return np.array(xc, yc, dtype=float)

    def get_drop_area(self):
        """
        Return the area of the 2D drop projection.
        """
        r = self.fits[1]
        xc, yc = self.fits[0]
        h = yc - self.baseline.get_projection_to_baseline([xc, yc])[1]
        theta = 2*(np.pi - np.arccos(h/r))
        # from wikipedia: circular segments
        area = r**2/2*(theta - np.sin(theta))
        return area

    def get_drop_volume(self):
        """
        Return the drop volume.
        """
        # From worlfram alpha on spehrical caps
        xyc, R = self.fits
        h = self.get_drop_height()
        V = 1/3*np.pi*h**2*(3*R - h)
        return V

    def get_fit_as_points(self, resolution=100):
        """
        Return a representation of the fit as point coordinates.
        """
        (xc, yc) = self.fits[0]
        radius = self.fits[1]
        theta = np.linspace(0, np.pi*2, resolution)
        x = xc + radius*np.cos(theta)
        y = yc + radius*np.sin(theta)
        pts = [x, y]
        return pts

    def display(self, displ_fits=True, displ_ca=True, displ_center=True,
                *args, **kwargs):
        """
        """
        super().display()
        # Display fits
        if self.fits is not None and displ_center:
            xyc, R = self.fits
            plt.plot(xyc[0], xyc[1], marker='o',
                     color=self.colors[5])
        if self.fits is not None and displ_fits:
            xyc, R = self.fits
            circ = plt.Circle((xyc[0], xyc[1]), radius=R,
                              color=self.colors[5],
                              fill=False)
            plt.gca().add_artist(circ)
        # Display contact angles
        if displ_ca:
            lines = self._get_angle_display_lines()
            for line in lines:
                plt.plot(line[0], line[1],
                         color=self.colors[0])
                plt.plot(line[0][0], line[0][1],
                         color=self.colors[0])


class DropCirclesFit(DropFit):
    def __init__(self, xyc, Rs, baseline, triple_pts, x_bounds, y_bounds):
        super().__init__(baseline=baseline,
                         x_bounds=x_bounds,
                         y_bounds=y_bounds)
        self.fits = [[xyc[i], Rs[i]] for i in range(len(xyc))]
        self.triple_pts = triple_pts

    def _get_inters_base_fit(self):
        """
        """
        # getting intersection points
        # from: http://mathworld.wolfram.com/Circle-LineIntersection.html
        (xc, yc), r = self.fits[0]
        xbas1 = self.baseline.pt1[0] - xc
        xbas2 = self.baseline.pt2[0] - xc
        ybas1 = self.baseline.pt1[1] - yc
        ybas2 = self.baseline.pt2[1] - yc
        dx = xbas2 - xbas1
        dy = ybas2 - ybas1
        dr = (dx**2 + dy**2)**.5
        D = xbas1*ybas2 - xbas2*ybas1
        #
        x1 = (D*dy + np.sign(dy)*dx*(r**2*dr**2 - D**2)**.5)/dr**2
        x2 = (D*dy - np.sign(dy)*dx*(r**2*dr**2 - D**2)**.5)/dr**2
        y1 = (-D*dx + abs(dy)*(r**2*dr**2 - D**2)**.5)/dr**2
        y2 = (-D*dx - abs(dy)*(r**2*dr**2 - D**2)**.5)/dr**2
        #
        x1 += xc
        x2 += xc
        y1 += yc
        y2 += yc
        return [[x1, y1], [x2, y2]]

    def _compute_fitting_angle(self):
        thetas = []
        xyc = self.fits[0][0]
        sides = [1, -1]
        for i in range(2):
            xyc2 = self.fits[i+1][0]
            theta = sides[i]*np.pi/2 + np.arctan2(xyc[1] - xyc2[1],
                                                  xyc[0] - xyc2[0])
            theta = theta % (2*np.pi)
            thetas.append(theta/np.pi*180)
        return np.array(thetas)

    def compute_contact_angle(self, verbose=False):
        """
        Compute the contact angles.

        Returns
        =======
        thetas : 2x1 array of numbers
           Contact angles in °
        """
        bs_angle = self.baseline.tilt_angle*180/np.pi
        # Compute circle fits contact angles
        (xc, yc), R = self.fits[0]
        pts = self._get_inters_base_fit()
        thetas = []
        for pt in pts:
            theta = np.pi/2 + np.arctan((yc - pt[1])/(xc - pt[0]))
            thetas.append(theta)
        self.thetas = np.array(thetas)*180/np.pi
        # correct regarding baseline angle
        self.thetas -= bs_angle
        # Compute contact angles at triple point
        if self.triple_pts is not None:
            self.thetas_triple = self._compute_fitting_angle()
            # correct regardin baseline angle
            self.thetas_triple -= bs_angle
        # display if asked
        if verbose:
            self.display()

    def get_drop_center(self):
        """
        Return the center of the drop.
        """
        return np.array(self.fits[0][0],
                        dtype=float)

    def get_drop_height(self):
        """
        Return the drop height.
        """
        hb = self.baseline.get_projection_to_baseline(self.fits[0][0])[1]
        return (self.fits[0][0][1] - hb) + self.fits[0][1]

    def get_fit_as_points(self, resolution=100):
        """
        Return a representation of the fit as point coordinates.
        """
        (xc, yc), radius = self.fits[0]
        (xc1, yc1), radius1 = self.fits[1]
        (xc2, yc2), radius2 = self.fits[2]
        theta = np.linspace(0, np.pi*2, resolution)
        # main circle
        x = xc + radius*np.cos(theta)
        y = yc + radius*np.sin(theta)
        # small circle 1
        x1 = xc1 + radius1*np.cos(theta)
        y1 = yc1 + radius1*np.sin(theta)
        # small circle 1
        x2 = xc2 + radius2*np.cos(theta)
        y2 = yc2 + radius2*np.sin(theta)
        # returning
        pts = [np.concatenate([x, [np.nan], x1, [np.nan], x2]),
               np.concatenate([y, [np.nan], y1, [np.nan], y2])]
        return np.array(pts)

    def display(self, displ_fits=True, displ_ca=True,
                displ_tp=True, *args, **kwargs):
        """
        """
        super().display()
        # Display fits
        if self.fits is not None and displ_fits:
            for cf in self.fits:
                xyc, R = cf
                plt.plot(xyc[0], xyc[1], marker='o',
                         color=self.colors[5])
                circ = plt.Circle((xyc[0], xyc[1]), radius=R,
                                  color=self.colors[5],
                                  fill=False)
                plt.gca().add_artist(circ)
        # Display contact angles
        if displ_ca:
            lines = self._get_angle_display_lines()
            for line in lines:
                plt.plot(line[0], line[1],
                         color=self.colors[0])
                plt.plot(line[0][0], line[0][1],
                         color=self.colors[0])
        # Display triple points from circle fits
        if self.triple_pts is not None and displ_tp:
            for tp in self.triple_pts:
                plt.plot(tp[0], tp[1], marker='o',
                         color=self.colors[5])


class DropEllipseFit(DropFit):
    def __init__(self, xyc, R1, R2, theta, baseline, x_bounds, y_bounds):
        super().__init__(baseline, x_bounds, y_bounds)
        self.fits = [xyc, R1, R2, theta]

    def _get_inters_base_fit(self):
        """
        """
        (h, k), a, b, theta = self.fits
        if np.any(np.isnan([h, k, a, b, theta])):
            return [[np.nan, np.nan], [np.nan, np.nan]]
        # Rotate the baseline in the ellipse referential
        tmpbs = self.baseline.copy()
        tmpbs.set_origin(h, k)
        tmpbs.rotate(-theta)
        tmpbs.set_origin(-h, -k)
        m, c = tmpbs.coefs
        # from: http://ambrsoft.com/TrigoCalc/Circles2/Ellipse/EllipseLine.htm
        eps = c - k
        delta = c + m*h
        # x
        A = h*b**2 - m*a**2*eps
        B = a*b*(a**2*m**2 + b**2 - delta**2 - k**2 + 2*delta*k)**.5
        C = a**2*m**2 + b**2
        x1 = (A + B)/C
        x2 = (A - B)/C
        # y
        D = b**2*delta + k*a**2*m**2
        E = a*b*m*(a**2*m**2 + b**2 - delta**2 - k**2 + 2*delta*k)**.5
        F = C
        y1 = (D + E)/F
        y2 = (D - E)/F
        # Rotate back the point in the base referential
        xs = [x1, x2]
        ys = [y1, y2]
        x1, x2 = h + np.cos(theta)*(xs - h) - np.sin(theta)*(ys - k)
        y1, y2 = k + np.sin(theta)*(xs - h) + np.cos(theta)*(ys - k)
        if x1 < x2:
            return [[x1, y1], [x2, y2]]
        else:
            return [[x2, y2], [x1, y1]]

    def compute_contact_angle(self, verbose=False):
        """
        Compute the contact angles.

        Returns
        =======
        thetas : 2x1 array of numbers
           Contact angles in °
        """
        bs_angle = self.baseline.tilt_angle*180/np.pi
        # Compute ellipse fit contact angle
        xy_inter = self._get_inters_base_fit()
        bs_angle = self.baseline.tilt_angle
        (xc, yc), R1, R2, theta = self.fits
        thetas = []
        for i, xy in enumerate(xy_inter):
            x_ref = (xy[0] - xc)*np.cos(theta) + (xy[1] - yc)*np.sin(theta)
            y_ref = (xy[1] - yc)*np.cos(theta) - (xy[0] - xc)*np.sin(theta)
            slope = -(R2**2*x_ref)/(R1**2*y_ref)
            thet = np.arctan(slope)
            if np.sin(thet) < 0:
                thet += np.pi
            thet += (theta - bs_angle)
            thetas.append(thet)
        self.thetas = np.array(thetas)*180/np.pi
        # display if asked
        if verbose:
            self.display()

    def get_drop_center(self):
        """
        Return the center of the drop.
        """
        return np.array(self.fits[0], dtype=float)

    def get_drop_height(self):
        """
        Return the drop height.
        """
        hb = self.baseline.get_projection_to_baseline(self.fits[0])[1]
        xyc, R1, R2, theta = self.fits
        R = ((R1*np.sin(theta))**2 + (R2*np.cos(theta))**2)**.5
        return (xyc[1] - hb) + R

    def get_fit_as_points(self, resolution=100):
        """
        Return a representation of the fit as point coordinates.
        """
        (xc, yc), R1, R2, theta = self.fits
        xs, ys = helpers.get_ellipse_points(xc, yc, R1, R2, theta, res=resolution)
        pts = [xs, ys]
        return pts

    def display(self, displ_fits=True, displ_ca=True,
                *args, **kwargs):
        """
        """
        super().display()
        # Display fits
        if displ_fits:
            (xc, yc), R1, R2, theta = self.fits
            elxs, elys = helpers.get_ellipse_points(xc, yc, R1, R2, theta, res=100)
            rxs = [xc + R1*np.cos(theta),
                   xc,
                   xc + R2*np.cos(theta + np.pi/2)]
            rys = [yc + R1*np.sin(theta),
                   yc,
                   yc + R2*np.sin(theta + np.pi/2)]
            plt.plot(rxs, rys, color=self.colors[3], ls=":")
            plt.plot(elxs, elys, color=self.colors[3])
            plt.plot(xc, yc, color=self.colors[3], marker='o', ls='none')
        # Display contact angles
        if displ_ca:
            lines = self._get_angle_display_lines()
            for line in lines:
                plt.plot(line[0], line[1],
                         color=self.colors[0])
                plt.plot(line[0][0], line[0][1],
                         color=self.colors[0])


class DropEllipsesFit(DropFit):
    def __init__(self, xyc1, R1a, R1b, theta1, xyc2, R2a, R2b, theta2,
                 baseline, x_bounds, y_bounds):
        super().__init__(baseline, x_bounds, y_bounds)
        self.fits = [[xyc1, R1a, R1b, theta1],
                     [xyc2, R2a, R2b, theta2]]

    def _get_inters_base_ellipse(self, fit):
        """
        """
        (h, k), a, b, theta = fit
        if np.any(np.isnan([h, k, a, b, theta])):
            return [[np.nan, np.nan], [np.nan, np.nan]]
        # Rotate the baseline in the ellipse referential
        tmpbs = self.baseline.copy()
        tmpbs.set_origin(h, k)
        tmpbs.rotate(-theta)
        tmpbs.set_origin(-h, -k)
        m, c = tmpbs.coefs
        # from: http://ambrsoft.com/TrigoCalc/Circles2/Ellipse/EllipseLine.htm
        eps = c - k
        delta = c + m*h
        # x
        A = h*b**2 - m*a**2*eps
        B = a*b*(a**2*m**2 + b**2 - delta**2 - k**2 + 2*delta*k)**.5
        C = a**2*m**2 + b**2
        x1 = (A + B)/C
        x2 = (A - B)/C
        # y
        D = b**2*delta + k*a**2*m**2
        E = a*b*m*(a**2*m**2 + b**2 - delta**2 - k**2 + 2*delta*k)**.5
        F = C
        y1 = (D + E)/F
        y2 = (D - E)/F
        # Rotate back the point in the base referential
        xs = [x1, x2]
        ys = [y1, y2]
        x1, x2 = h + np.cos(theta)*(xs - h) - np.sin(theta)*(ys - k)
        y1, y2 = k + np.sin(theta)*(xs - h) + np.cos(theta)*(ys - k)
        if x1 < x2:
            return [[x1, y1], [x2, y2]]
        else:
            return [[x2, y2], [x1, y1]]

    def _get_inters_base_fit(self):
        inter1 = self._get_inters_base_ellipse(self.fits[0])
        inter2 = self._get_inters_base_ellipse(self.fits[1])
        return [inter1[0], inter2[1]]

    def compute_contact_angle(self, verbose=False):
        """
        Compute the contact angles.

        Returns
        =======
        thetas : 2x1 array of numbers
           Contact angles in °
        """
        bs_angle = self.baseline.tilt_angle*180/np.pi
        # Compute ellipse fit contact angle
        xy_inter = self._get_inters_base_fit()
        bs_angle = self.baseline.tilt_angle
        thetas = []
        for i, xy in enumerate(xy_inter):
            (xc, yc), R1, R2, theta = self.fits[i]
            x_ref = (xy[0] - xc)*np.cos(theta) + (xy[1] - yc)*np.sin(theta)
            y_ref = (xy[1] - yc)*np.cos(theta) - (xy[0] - xc)*np.sin(theta)
            slope = -(R2**2*x_ref)/(R1**2*y_ref)
            thet = np.arctan(slope)
            if np.sin(thet) < 0:
                thet += np.pi
            thet += (theta - bs_angle)
            thetas.append(thet)
        self.thetas = np.array(thetas)*180/np.pi
        # display if asked
        if verbose:
            self.display()

    def get_drop_center(self):
        """
        Return the center of the drop.
        """
        return np.array([(self.fits[0][0] + self.fits[1][0])/2,
                         (self.fits[0][1] + self.fits[1][1])/2],
                        dtype=float)

    def get_fit_as_points(self, resolution=100):
        """
        Return a representation of the fit as point coordinates.
        """
        inters = self._get_inters_base_fit()
        xs = []
        ys = []
        for i in range(2):
            (xc, yc), R1, R2, theta = self.fits[i]
            txs, tys = helpers.get_ellipse_points(xc, yc, R1, R2, theta,
                                                  res=resolution)
            # remove useless part
            if i == 0:
                filt = np.logical_and(txs < xc,
                                      tys > inters[i][1])
            else:
                filt = np.logical_and(txs > xc,
                                      tys > inters[i][1])
            txs = txs[filt]
            tys = tys[filt]
            # Nothing remaining
            if len(txs) == 0:
                txs = [np.nan]
                tys = [np.nan]
                xs.append(txs)
                ys.append(tys)
                continue
            # Make sure it starts at the intersection point
            ind = np.argmax(abs(tys - np.roll(tys, 1)))
            txs = np.roll(txs, -ind)
            tys = np.roll(tys, -ind)
            #
            xs.append(txs)
            ys.append(tys)
        # concatenate
        if i == 0:
            xs = np.concatenate((xs[0], [np.nan], xs[1], [np.nan]))
            ys = np.concatenate((ys[0], [np.nan], ys[1], [np.nan]))
        else:
            xs = np.concatenate((xs[0], [np.nan], xs[1][::-1], [np.nan]))
            ys = np.concatenate((ys[0], [np.nan], ys[1][::-1], [np.nan]))
        pts = [xs, ys]
        return pts

    def display(self, displ_fits=True, displ_ca=True,
                *args, **kwargs):
        """
        """
        super().display()
        # Display fits
        if displ_fits:
            for i in range(2):
                (xc, yc), R1, R2, theta = self.fits[i]
                elxs, elys = helpers.get_ellipse_points(xc, yc, R1, R2, theta,
                                                        res=100)
                # Filter out wrong sides
                if i == 0:
                    filt = elxs < xc
                else:
                    filt = elxs > xc
                elxs = elxs[filt]
                elys = elys[filt]
                # be sure to display ellipse radius on the right side
                coef = 1
                if np.cos(theta) > 0 and i == 0:
                    coef = -1
                if np.cos(theta) < 0 and i == 1:
                    coef = -1
                # plot
                rxs = [xc + coef*abs(R1*np.cos(theta)),
                       xc,
                       xc + coef*R2*np.cos(theta + np.pi/2)]
                rys = [yc + coef*R1*np.sin(theta),
                       yc,
                       yc + coef*R2*np.sin(theta + np.pi/2)]
                plt.plot(rxs, rys, color=self.colors[3], ls=":")
                plt.plot(elxs, elys, color=self.colors[3])
                plt.plot(xc, yc, color=self.colors[3], marker='o', ls='none')
        # Display contact angles
        if displ_ca:
            lines = self._get_angle_display_lines()
            for line in lines:
                plt.plot(line[0], line[1],
                         color=self.colors[0])
                plt.plot(line[0][0], line[0][1],
                         color=self.colors[0])



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2018 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://github.com/gabylaunay/pyDSA_core

# This file is part of pyDSA_core

# pyDSA_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import numpy as np
import json
import os
import warnings
from IMTreatment import TemporalScalarFields
from IMTreatment.utils import ProgressCounter
from IMTreatment.stabilize import Stabilizer
import IMTreatment.plotlib as pplt
from . import image
from . import baseline
from . import temporaldropedges as tde
from . import dropedges


"""  """

__author__ = "Gaby Launay"
__copyright__ = "Gaby Launay 2017"
__credits__ = ""
__email__ = "gaby.launay@tutanota.com"
__status__ = "Development"


class TemporalImages(TemporalScalarFields):
    def __init__(self, filepath=None, cache_infos=True):
        super().__init__()
        self.baseline = None
        self.field_type = image.Image
        if filepath is not None:
            self.filepath = os.path.abspath(filepath)
            if os.path.isdir(self.filepath):
                self.infofile_path = os.path.join(self.filepath, "info.info")
            else:
                self.infofile_path = os.path.splitext(self.filepath)[0] + ".info"
            self.cache_infos = cache_infos
        else:
            self.filepath = None
            self.infofile_path = None
            self.cache_infos = False

    def add_field(self, field, time=0, unit_times="", copy=True):
        super().add_field(field=field, time=time, unit_times=unit_times,
                          copy=copy)
        if self.baseline:
            self.fields[-1].baseline = self.baseline
        else:
            self.baseline = self.fields[-1].baseline

    def set_baseline(self, pt1, pt2):
        """
        Set the drop baseline.
        """
        for i in range(len(self.fields)):
            self.fields[i].set_baseline(pt1=pt1, pt2=pt2)
        self.baseline = baseline.Baseline(
            [pt1, pt2],
            xmax=self.axe_x[-1],
            xmin=self.axe_x[0])
        if self.cache_infos:
            self._dump_infos()

    def set_evolving_baseline(self, baseline1, baseline2):
        """
        Set a linearly evolving baseline.
        """
        base1_pt1 = np.asarray(baseline1[0])
        base1_pt2 = np.asarray(baseline1[1])
        base2_pt1 = np.asarray(baseline2[0])
        base2_pt2 = np.asarray(baseline2[1])
        imax = len(self.fields) - 1
        for i in range(len(self.fields)):
            ratio = i/imax
            pt1 = base1_pt1*(1 - ratio) + base2_pt1*ratio
            pt2 = base1_pt2*(1 - ratio) + base2_pt2*ratio
            self.fields[i].set_baseline(pt1, pt2)
        self.baseline = "evolving"
        if self.cache_infos:
            self._dump_infos()

    def set_origin(self, x, y):
        """
        Modify the axis in order to place the origin at the given point (x, y)

        Parameters
        ----------
        x : number
        y : number
        """
        for im in self.fields:
            im.set_origin(x, y)
        self.axe_x = self.fields[0].axe_x
        self.axe_y = self.fields[0].axe_y
        self.baseline = self.fields[0].baseline
        if self.cache_infos:
            self._dump_infos()

    def choose_baseline(self, ind_image=None):
        """
        Choose baseline position interactively.

        Select as many points as you want (click on points to delete them),
        and close the figure when done.
        """
        #
        if ind_image is None:
            ind_image = int(len(self.fields)/2)
        self.baseline = self.fields[ind_image].choose_baseline()
        # No baseline set
        if self.baseline is None:
            return None
        for i in range(len(self.fields)):
            self.fields[i].baseline = self.baseline
        if self.cache_infos:
            self._dump_infos()
        return self.baseline

    def track_baseline(self, ind_image=0, orb_kwargs={}, mode='continuous',
                       verbose=False, dense=False):
        """
        Track the baseline from the wanted frame.

        Use opencv features tracking algorithm.
        """
        stab = Stabilizer(self, orb_kwargs=orb_kwargs, mode=mode)
        stab._compute_transform(dense=dense, opt_flow_args={}, verbose=verbose)
        stab._smooth_transform(smooth_size=0, verbose=verbose)
        pt1 = self[0].baseline.pt1
        pt2 = self[0].baseline.pt2
        pts = stab._apply_transform_to_point([pt1, pt2], from_frame=ind_image,
                                             verbose=verbose)
        # apply new baselines
        for i, im in enumerate(self.fields):
            im.set_baseline(pt1=pts[0][i], pt2=pts[1][i])
        # set baseline to evolving
        self.baseline = "evolving"
        # return
        return stab

    def display(self, *args, **kwargs):
        super().display(*args, **kwargs)
        if isinstance(self.baseline, baseline.Baseline):
            self.baseline.display()
        elif self.baseline == "evolving":
            # Display evolving baseline
            xs = []
            ys = []
            for field in self.fields:
                bl = field.baseline
                xs.append([bl.pt1[0], bl.pt2[0]])
                ys.append([bl.pt1[1], bl.pt2[1]])
            db = pplt.Displayer(xs, ys, color=self[0].colors[0])
            pplt.ButtonManager(db)
        else:
            pass

    def scale_interactive(self, indice=None):
        """
        Scale the Images interactively.

        Parameters
        ==========
        indice: integer
            Frame number on which do the scaling.
            (default to the mid one).
        """
        indice = indice or int(len(self.fields)/2)
        scale, unit = self.fields[indice].scale_interactive()
        if scale is None and unit is None:
            return None
        self.scale(scalex=scale, scaley=scale, inplace=True)
        self.unit_x = unit
        self.unit_y = unit
        if self.cache_infos:
            self._dump_infos()

    def scale(self, scalex=None, scaley=None, scalev=None, scalet=None,
              inplace=False):
        """
        Scale the Fields.

        Parameters
        ----------
        scalex, scaley, scalev : numbers or Unum objects
            Scale for the axis and the values.
        inplace : boolean
            .
        """
        if inplace:
            tmp_f = self
        else:
            tmp_f = self.copy()
        old_dx = self.dx
        old_dy = self.dy
        # scale the scalarfields
        super(TemporalImages, tmp_f).scale(scalex=scalex, scaley=scaley,
                                           scalev=scalev, scalet=scalet,
                                           inplace=True)
        new_dx = self.dx
        new_dy = self.dy
        if self.baseline is not None:
            tmp_f.baseline.scale(scalex=new_dx/old_dx,
                                 scaley=new_dy/old_dy)
            for f in self.fields:
                f.baseline = tmp_f.baseline
        return tmp_f

    def _dump_infos(self):
        # Gather old information if necessary
        if os.path.isfile(self.infofile_path):
            with open(self.infofile_path, 'r') as f:
                dic = json.load(f)
        else:
            dic = {}
        # Update with new values
        unit_x = self.unit_x.strUnit()[1:-1]
        unit_y = self.unit_y.strUnit()[1:-1]
        if self.baseline != "evolving" and self.baseline is not None:
            pt1 = list(self.baseline.pt1)
            pt2 = list(self.baseline.pt2)
        else:
            pt1 = None
            pt2 = None
        new_dic = {"dx": self.dx,
                   "dy": self.dy,
                   "x0": self.axe_x[0],
                   "y0": self.axe_y[0],
                   "baseline_pt1": pt1,
                   "baseline_pt2": pt2,
                   "unit_x": unit_x,
                   "unit_y": unit_y}
        dic.update(new_dic)
        # Write back infos
        with open(self.infofile_path, 'w+') as f:
            json.dump(dic, f)

    def _import_infos(self):
        if not os.path.isfile(self.infofile_path):
            return None
        try:
            with open(self.infofile_path, 'r') as f:
                dic = json.load(f)
        except:
            warnings.warn('Corrupted infofile, reinitializing...')
            os.remove(self.infofile_path)
            return None
        # Update with infos
        dx = dic['dx']
        dy = dic['dy']
        self.scale(scalex=dx/(self.axe_x[1] - self.axe_x[0]),
                   scaley=dy/(self.axe_y[1] - self.axe_y[0]),
                   inplace=True)
        if self.unit_x.strUnit()[1:-1] != dic['unit_x']:
            self.unit_x = dic['unit_x']
        if self.unit_y.strUnit()[1:-1] != dic['unit_y']:
            self.unit_y = dic['unit_y']
        try:  # Compatibility
            x0 = dic['x0']
            y0 = dic['y0']
            self.axe_x += (x0 - self.axe_x[0])
            self.axe_y += (y0 - self.axe_y[0])
        except KeyError:
            pass
        base1 = dic['baseline_pt1']
        base2 = dic['baseline_pt2']
        if base1 is not None and base2 is not None:
            self.set_baseline(base1, base2)

    def edge_detection_canny(self, threshold1=None, threshold2=None,
                             base_max_dist=30, size_ratio=.5,
                             ignored_pixels=2, nmb_edges=2,
                             keep_exterior=True,
                             dilatation_steps=1,
                             smooth_size=None,
                             iteration_hook=None,
                             verbose=False):
        """
        Perform edge detection.

        Parameters
        ==========
        threshold1, threshold2: integers
            Thresholds for the Canny edge detection method.
            (By default, inferred from the data histogram)
        base_max_dist: integers
            Maximal distance (in pixel) between the baseline and
            the beginning of the drop edge (default to 30).
        nmb_edges: integer
            Number of maximum expected edges (default to 2).
        size_ratio: number
            Minimum size of edges, regarding the bigger detected one
            (default to 0.5).
        keep_exterior: boolean
            If True (default), only keep the exterior edges.
        smooth_size: number
            If specified, the image is smoothed before
            performing the edge detection.
            (can be useful to put this to 1 to get rid of compression
             artefacts on images).
        dilatation_steps: positive integer
            Number of dilatation/erosion steps.
            Increase this if the drop edges are discontinuous.
        iteration_hook: function
            Hook run at each iterations with the iteration number
            and the total number of iterations planned.
        """
        # check
        if self.baseline is None:
            raise Exception('You have to define a baseline first.')
        #
        all_edge_empty = True
        pts = tde.TemporalDropEdges()
        pts.baseline = self.baseline
        if verbose:
            pg = ProgressCounter(init_mess="Detecting drop edges",
                                 nmb_max=len(self.fields),
                                 name_things='images',
                                 perc_interv=5)
        for i in range(len(self.fields)):
            try:
                pt = self.fields[i].edge_detection_canny(
                    threshold1=threshold1,
                    threshold2=threshold2,
                    base_max_dist=base_max_dist,
                    size_ratio=size_ratio,
                    ignored_pixels=ignored_pixels,
                    keep_exterior=keep_exterior,
                    nmb_edges=nmb_edges,
                    dilatation_steps=dilatation_steps,
                    smooth_size=smooth_size)
                all_edge_empty = False
            except Exception:
                pt = dropedges.DropEdges(xy=[], im=self, type='canny')
            pts.add_pts(pt, time=self.times[i], unit_times=self.unit_times)
            if iteration_hook is not None:
                iteration_hook(i, len(self.fields))
            if verbose:
                pg.print_progress()
        # check if edges has been detected
        if all_edge_empty:
            raise Exception('No edges could be detected. You should'
                            ' check the baseline position.')
        return pts

    def edge_detection_contour(self, size_ratio=.5, nmb_edges=2,
                               level=0.5, ignored_pixels=2,
                               iteration_hook=None,
                               verbose=False):
        """
        Perform edge detection.

        Parameters
        ==========
        nmb_edges: integer
            Number of maximum expected edges (default to 2).
        level: number
            Normalized level of the drop contour.
            Should be between 0 (black) and 1 (white).
            Default to 0.5.
        size_ratio: number
            Minimum size of edges, regarding the bigger detected one
            (default to 0.5).
        ignored_pixels: integer
            Number of pixels ignored around the baseline
            (default to 2).
            Putting a small value to this allow to avoid
            small surface defects to be taken into account.
        """
        pts = tde.TemporalDropEdges()
        pts.baseline = self.baseline
        if verbose:
            pg = ProgressCounter("Detecting drop edges", "Done",
                                 len(self.fields), 'images', 5)
        for i in range(len(self.fields)):
            try:
                pt = self.fields[i].edge_detection_contour(nmb_edges=nmb_edges,
                                                           level=level,
                                                           ignored_pixels=ignored_pixels,
                                                           size_ratio=size_ratio)
            except Exception:
                pt = dropedges.DropEdges(xy=[], im=self, type='contour')
            pts.add_pts(pt, time=self.times[i], unit_times=self.unit_times)
            if verbose:
                pg.print_progress()
            if iteration_hook is not None:
                iteration_hook(i, len(self.fields))
        return pts

    edge_detection = edge_detection_canny
